local lgr = love.graphics
local win = love.window
local lfs = love.filesystem
local kbd = love.keyboard
local tmr = love.timer

platform = {}

function platform:create(v)
	local _i = {}
	setmetatable(_i, {__index = self})
	_i:setup(v or {})
	return _i
end
function platform:setup(v)
	self.world = v.world or game.map.collisionMap
	self.facing = 1
	self.x = v.x or 128
	self.y = v.y or 128
	self.w = v.w or 32
	self.h = v.h or 8
	self.originY = self.y
	self.xSpeed = v.xSpeed or 0
	self.ySpeed = v.ySpeed or 0
	self.kind = v.kind or "wall"
	self.map = v.map or game.map
	self.isPlatform = true
	self.kind = "wall"
	self.solidTop = v.solidTop or nil
	self.solid = v.solid or nil
	-- if not self.solidTop and not self.solid then self.solid = true end
	self.slippery = v.slippery or nil
	self.nameID = v.nameID
	self.path = v.path or nil
	self.pathStep = 0
	self.active = v.active or false
	self.activateOnTouch = v.activateOnTouch or nil
	self.sinkWhenHeavy = v.sinkWhenHeavy or false
	self.sinkTolerance = v.sinkTolerance or 4
end

function platform:update(dt)
	local world = self.world

	local xs, ys = 0, 0
	if self.active then
		xs, ys = self.xSpeed, self.ySpeed
	end

	if self.sinkWhenHeavy then
		if game.player.attachedToPlatform == self.nameID then
			ys = ys + 20
		else
			ys = ys - 20
		end
	end

	local future_x = self.x + xs * dt
	local future_y = self.y + ys * dt

	if future_y > self.originY + self.sinkTolerance then
		future_y = self.originY + self.sinkTolerance
	end

	if future_y < self.originY then
		future_y = self.originY
	end

	local func
	if self.solid then
		func = function(other)
			if other.kind == "player" then
				return "touch"
			end
			return false
		end
	else
		func = function() return false end
	end

	local next_x, next_y, cols, len = world:move(self, future_x, future_y, func)

	for i=1, len do
		local col = cols[i]
		local other = col.other
		if col.normal.x ~= 0 then
			if other.kind == "player" then
				game.player:moveWithPlatform(dt, xs, ys)
				-- print("DD", tmr.getTime())
			end
		end
	end

	self.x, self.y = next_x, next_y

	if game.player.attachedToPlatform == self.nameID then
		game.player:moveWithPlatform(dt, xs, ys)
	end
end

function platform:draw()
	if self.path then
		for i = 1, #self.path do
			local p = self.path[i]
			lgr.push()
			lgr.translate(math.floor(p[1]) * 16, math.floor(p[2]) * 16)

			lgr.setColor(0,0,255)
			lgr.circle("fill", 0, 0, 2)

			lgr.pop()
		end
	end

	lgr.push()
	lgr.translate(math.floor(self.x), math.floor(self.y))

	lgr.setColor(255,200,0)
	lgr.rectangle("fill", 0, 0, self.w, self.h)
	lgr.setColor(0,0,0,100)
	lgr.rectangle("fill", 0, 0, self.w, 2)

	lgr.pop()
end

function platform:destroy()
  self.world:remove(self)
end


