local lgr = love.graphics
local win = love.window
local lfs = love.filesystem
local kbd = love.keyboard
local tmr = love.timer

smbPlayer = {}

function smbPlayer:create(v)
	local _i = {}
	setmetatable(_i, {__index = self})
	_i:setup(v or {})
	return _i
end
function smbPlayer:setup(v)
	self.world = v.world
	self.facing = 1
	self.x = v.x or 2 * 16
	self.y = v.y or 2 * 16
	self.w = v.w or width
	self.h = v.h or height
	self.vx = 0
	self.vy = 0
	self.iox = v.iox or 0
	self.ioy = v.ioy or 0
	self.isDead = false
	self.map = v.map or game.map
	self.gravityAccel  = 600 -- pixels per second^2
	self.deadDuration  = 3   -- seconds until res-pawn
	self.runAccel      = 300 -- the player acceleration while going left/right
	self.brakeAccel    = 500
	self.iceAccel	   = 100
	self.iceBrake	   = 160
	self.waterAccel	   = 40
	self.waterBrake	   = 100
	self.runSpeed      = 96
	self.jumpVelocity  = 300 -- the initial upwards velocity when jumping
	self.jumpHeightMax = 48
	self.jumpHeight = 0
	self.ladderClimbSpeed = 64
	self.jumpCooldown = 0
	self.kind = "player"
	self.id = "player"
	self.pState = v.pState
	self.attachedToPlatform = ""
	self.health = 10000
	self.invTime = 0

	self.leftWater = true
	self._wasInWater = false

	self.playerImage = lgr.newImage("images/player.png")
	self.playerQuad = {}
	for y = 0, 2 do
		self.playerQuad["walk_"..(y)] = lgr.newQuad(0, y * 32, 16, 32, self.playerImage:getWidth(), self.playerImage:getHeight())
		self.playerQuad["walkcarry_"..(y)] = lgr.newQuad(16, y * 32, 16, 32, self.playerImage:getWidth(), self.playerImage:getHeight())
	end
	for x = 0, 1 do
		self.playerQuad["climb_"..(x)] = lgr.newQuad(32 + x * 16, 32, 16, 32, self.playerImage:getWidth(), self.playerImage:getHeight())
	end

	self.playerQuad["pickup"] = lgr.newQuad(32, 64, 16, 32, self.playerImage:getWidth(), self.playerImage:getHeight())
	self.playerQuad["throw"] = lgr.newQuad(48, 64, 16, 32, self.playerImage:getWidth(), self.playerImage:getHeight())


	self.pausingForAnimation = ""
	self.pauseTime = 0
	self.frame = 0
end

function smbPlayer:update(dt)
	self.jumpCooldown = math.max(self.jumpCooldown - dt, 0)

	self.invTime = math.max(self.invTime - dt, 0)
	if self.invTime <= 0 then

		if self.pauseTime > 0 then
			self.pauseTime = self.pauseTime - dt
			if self.pauseTime <= 0 then
				self.pausingForAnimation = ""
				self.pauseTime = 0
			end
		end

		if self.pauseTime == 0 then
			self:changeVelocityByKeys(dt)
			self:changeVelocityByGravity(dt)
		end

		self.watersensor = self.map.waterMap[math.ceil(((self.x + self.w / 2) / self.map.tileSize))][math.ceil((self.y + self.h / 3) / self.map.tileSize)] or 0

		self.chestl = self.map.grid[math.ceil((self.x) / self.map.tileSize)][math.ceil((self.y + self.h / 2) / self.map.tileSize)]
		self.chestc = self.map.grid[math.ceil(((self.x + self.w / 2) / self.map.tileSize))][math.ceil((self.y + self.h / 2) / self.map.tileSize)]
		self.chestr = self.map.grid[math.ceil((self.x + self.w) / self.map.tileSize)][math.ceil((self.y + self.h / 2) / self.map.tileSize)]

		self.feetl = self.map.grid[math.ceil((self.x) / self.map.tileSize)][math.ceil((self.y + self.h) / self.map.tileSize)]
		self.feetc = self.map.grid[math.ceil(((self.x + self.w / 2) / self.map.tileSize))][math.ceil((self.y + self.h) / self.map.tileSize)]
		self.feetr = self.map.grid[math.ceil((self.x + self.w) / self.map.tileSize)][math.ceil((self.y + self.h) / self.map.tileSize)]

		self.tlb = self.map.grid[math.ceil((self.x) / self.map.tileSize)][math.ceil((self.y + self.h + 1) / self.map.tileSize)]
		self.tcb = self.map.grid[math.ceil(((self.x + self.w / 2) / self.map.tileSize))][math.ceil((self.y + self.h + 1) / self.map.tileSize)]
		self.trb = self.map.grid[math.ceil((self.x + self.w) / self.map.tileSize)][math.ceil((self.y + self.h + 1) / self.map.tileSize)]

		self.standingOnBlock = self.map.grid[math.ceil((self.x + (self.w / 2)) / self.map.tileSize)][math.ceil((self.y + self.h) / self.map.tileSize) + 1]
		self.inWater = (self.watersensor > 0)

		if self.inWater ~= self._wasInWater then
			if self.inWater then
				-- Enter the water
				if self.leftWater then
					self:createSplashParticles()
				end
				self.vy = self.vy / 10
				self.leftWater = false
			else
				-- Exit the water
				if kbd.isDown("up") then
					self:createSplashParticles()
					self:jump(true)
					self.leftWater = true
				end
			end
		end

		if kbd.isDown("lshift") then
			self.runSpeed = 134
		else
			self.runSpeed = 96
		end

		local _runSpeed = self.runSpeed
		if self.inWater then _runSpeed = 70 end

		self.vy = math.max(math.min(self.vy, 255), -255)
		self.vx = math.max(math.min(self.vx, _runSpeed), -_runSpeed)

		if self.vy < 0 then
			self.movingVert = -1
			self.jumpHeight = self.jumpHeight + dt * math.abs(self.jumpVelocity)

			if self.jumpHeight > self.jumpHeightMax then
				self.vy = 0
			end
		elseif self.vy > 0 then
			self.movingVert = 1
		else
			self.movingVert = 0
		end

		self:moveColliding(dt)
		self:changeVelocityByBeingOnGround(dt)

		self._wasInWater = self.inWater

		if self.y > (self.map.height + 2) * self.map.tileSize then
			-- local nx, ny = 2 * self.map.tileSize + 2, 2 * self.map.tileSize
			self.vy = 0
			self.vx = 0
			game:createPlayer()
		end

		if self.onLadder and self.moving then
			self.frame = self.frame + dt * 6
			if self.frame >= 2 then self.frame = math.max(self.frame - 2, 0) end
		elseif self.vx ~= 0 then
			local s = (10 * (math.abs(self.vx) / 100))
			if self.onIce then
				s = (20 * (math.abs(self.vx) / 100))
			end
			self.frame = self.frame + dt * s
			if self.frame >= 3 then self.frame = math.max(self.frame - 3, 0) end
		else
			self.frame = 0
		end

	end
end

function smbPlayer:draw()
	local f = math.floor(self.frame)
	local sprite = "walk_"..f

	-- lgr.setColor(255,100,255)
	-- lgr.rectangle("fill", math.floor(self.x), math.floor(self.y), self.w, self.h)
	-- if self.onLadder then
	-- 	lgr.setColor(0,244,255,200)
	-- else
	-- 	lgr.setColor(0,0,255,200)
	-- end
	-- lgr.rectangle("fill", math.floor(self.x + 1), math.floor(self.y + 1), self.w - 2, self.h - 2)

	if self.carryingBlockType then
		sprite = "walkcarry_"..f
		if self.carryingBlockType == "block" then
			lgr.push()
			lgr.translate(math.floor(self.x - 2), math.floor(self.y - 16))
			lgr.setColor(100,200,255)
			lgr.rectangle("fill", 0, 0, 16, 16 / 2)
			lgr.setColor(0,255,255)
			lgr.rectangle("fill", 0 + 2, 0 + 16 / 2, 16 - 4, 16 / 2)
			lgr.pop()
		else
			lgr.push()
			lgr.translate(math.floor(self.x + 2), math.floor(self.y - 8))
			lgr.setColor(100,200,255)
			lgr.rectangle("fill", 0, 0, 8, 8)
			lgr.pop()
		end
	end

	if self.isJumping or self.vy ~= 0 then
		sprite = "walkcarry_2"
	end

	if self.onLadder then
		sprite = "climb_"..f
	end

	if not self.playerQuad[sprite] then
		sprite = "walk_0"
	end

	if self.pausingForAnimation == "pickup" then
		sprite = "pickup"
	end

	lgr.setColor(255,255,255)
	lgr.draw(self.playerImage, self.playerQuad[sprite], math.floor(self.x + self.w / 2), math.floor(self.y), 0, -self.facing, 1, 8, self.ioy)

end

function smbPlayer:keypressed(k, r)
	if k == " " then
		self:pickup()
	elseif k == keyJump then
		self:jump()
	end
end

function smbPlayer:createSplashParticles()
	for i = 0, 15, 2 do
		local n = "particle_" .. math.random(1,10000)
		game.projectiles[n] = smbParticle:create { x = self.x + i, y = self.y + (self.h / 3) + 3, vx = (i * 4) - (8 * 4), vy = math.random(-30,-50), color = {0,100,255 - math.random(0,10),200}, life = 1, gravity = (math.random() * 50) + 50 }
	end
end

function smbPlayer:jump(w, m)
	if self.onGround or self.onLadder or self.inWater or w then
		local jumpVel
		if self.inWater then
			jumpVel = self.jumpVelocity / 5
		else
			jumpVel = self.jumpVelocity
		end
		self.onLadder = false
		self.vy = -jumpVel
		self.isJumping = true
		self.jumpHeight = 0
		self.jumpCooldown = 0.1
		self.jumpNow = false
	end
end

function smbPlayer:pickup()
	local sps = self.pState

	if self.carryingBlockType then
		-- Throw whatever is in your hand
		print(self.carryingBlockID, self.carryingBlockType)

		local vx = ((130 + math.abs(self.vx)) * self.facing)
		local vy = -(10 + math.abs(self.vy))
		if self.vx == 0 and self.carryingBlockType ~= "veggie" then
			vx = 10 * self.facing
		end
		if kbd.isDown("down") then
			vy = 30 + math.abs(self.vy)
		end
		local n = #sps.projectiles+1
		local w, h = 16, 16
		if self.carryingBlockType == "block" then
			w = 16 - 4
			h = 16 - 8
		else
			w = 8
			h = 8
		end
		local x
		if self.facing == -1 then
			x = self.x - (2 + 10)
		elseif self.facing == 1 then
			x = self.x - 2 + 10
		end
		sps.projectiles[n] = smbBlockProjectile:create {
			x = x, y = (self.y + self.h) - 32, w = w, h = h, vx = vx, vy = vy, isProjectile = true,
			world = sps.map.collisionMap, kind = self.carryingBlockType, id = self.carryingBlockID, solid = false, beingThrown = true }
		sps.map.collisionMap:add(sps.projectiles[n], sps.projectiles[n].x, sps.projectiles[n].y, sps.projectiles[n].w, sps.projectiles[n].h)
		self.carryingBlockID = nil
		self.carryingBlockType = nil
	elseif self.standingOnBlock == 3 and self.onGround then
		-- Check if standing on a pickupable block and pick it up
		sps.map.collisionMap:remove(sps.map.throwableBlockList[self.standingOnBlockID])
		sps.map.throwableBlockList[self.standingOnBlockID] = nil
		sps.map.grid[self.pickBlockX][self.pickBlockY] = 0

		sps:checkIfBlockNextToWater(self.pickBlockX, self.pickBlockY)

		self.carryingBlockID = self.standingOnBlockID
		self.carryingBlockType = "block"

		self.pausingForAnimation = "pickup"
		self.pauseTime = 0.1
		self.vx = 0

		print("Picked up Block")
	elseif self.standingOnBlock == 5 and self.onGround then
		local x, y = self.pickBlockX, self.pickBlockY

		for jx = 0, 15, 4 do
			for jy = 0, 15, 4 do
				local n = "particle_" .. math.random(1,10000)
				game.projectiles[n] = smbParticle:create { x = ((x-1) * 16) + jx + 2, y = ((y-1) * 16) + jy, vx = math.random(-2,2), vy = math.random(-40,-20),
															 color = {255-math.random(0,10),200,math.random() * 100}, life = 1, gravity = (math.random() * 50)+50, shape = "square" }
			end
		end

		if self.map.grid[x][y] == 5 then
			self.map.grid[x][y] = 0
			local n = "sand_" .. x .. "_" .. y
			sps.map.collisionMap:remove(sps.map.boxList[n])
			sps.map.boxList[n] = nil
		end

		self.pausingForAnimation = "pickup"
		self.pauseTime = 0.25
		self.vx = 0

		print("Picked up Block")
	elseif sps.map.plantGrid[self.pickBlockX][self.pickBlockY].item ~= "" and self.onGround then
		-- Check for a plant in the ground and pull it up if available
		local p = sps.map.plantGrid[self.pickBlockX][self.pickBlockY]
		print("Picked up a plant \"" .. p.item .. "\"")


		self.carryingBlockType = p.item

		self.pausingForAnimation = "pickup"
		self.pauseTime = 0.1
		self.vx = 0

		p.item = ""
	else
		print("No")
	end
end

function smbPlayer:changeVelocityByBeingOnGround(dt)
	if self.onGround then
		self.vy = math.min(self.vy, 0)
	end
end

function smbPlayer:checkIfOnGround(ny)
	self.isJumping = false
	if ny < 0 then
		self.onGround = true
	end
end

local playerFilter = function(other)
	local p = game.player
	local okind = other.kind

	if other.ladderTop and (p.onLadder or p.feetc == 9) then return false end
	-- if other.ladderTop and (p.onLadder or p.feetc == 9) then return "slide" end
	if not (other.solidTop and p.y + p.h > other.y) then return "slide" end
	if okind == "enemy" then return false end
	if okind == "bomb" then return false end
	if other.isProjectile then
		if (okind == "block" and not other.beingThrown) or not okind == "bomb" then return "slide" end
		return "cross"
	end
	if other.solid then return "slide" end
	return false
end

function smbPlayer:moveColliding(dt)
	self.onGround = false
	local world = self.world

	local future_x = self.x + self.vx * dt
	local future_y = self.y + self.vy * dt

	local next_x, next_y, cols, len = world:move(self, future_x, future_y, playerFilter)

	self.attachedToPlatform = ""
	for i=1, len do
		local col = cols[i]
		local other = col.other

		if col.normal.x ~= 0 then
			-- print(other.kind, tmr.getTime())
			if other.kind == "enemy" then
				print("DDSDSD")
			end
		end

		if col.normal.y == -1 then
			self.onIce = other.slippery or false

			if other.isPlatform then
				self.attachedToPlatform = other.nameID

				if not other.active and other.activateOnTouch then
					other.active = true
				end
			end

			if other.conveyorDirection then
				next_x = next_x + dt * (16 * other.conveyorDirection)
			end
			if self.standingOnBlock == 3 then
				self.standingOnBlockID = other.id or 0
			end
			self.pickBlockX = math.ceil((self.x + (self.w / 2)) / game.map.tileSize)
			self.pickBlockY = math.ceil((self.y + self.h) / game.map.tileSize) + 1
		end

		self:changeVelocityByCollisionNormal(col.normal.x, col.normal.y, bounciness)
		self:checkIfOnGround(col.normal.y)
	end

	self.x, self.y = next_x, next_y
end

function smbPlayer:moveWithPlatform(dt, vx, vy)
	local world = self.world

	local future_x = self.x + vx * dt
	local future_y = self.y + vy * dt

	local next_x, next_y, cols, len = world:move(self, future_x, future_y, playerFilter)

	self.x, self.y = next_x, next_y
end

function smbPlayer:changeVelocityByKeys(dt)
	if self.isDead then return end

	local vx, vy = self.vx, self.vy
	local brakeAccel, runAccel

	if self.inWater then
		brakeAccel = self.brakeAccel
		runAccel = self.runAccel
	elseif self.onIce then
		brakeAccel = self.iceBrake
		runAccel = self.iceAccel
	elseif self.onGround then
		brakeAccel = self.brakeAccel
		runAccel = self.runAccel
	else
		brakeAccel = self.brakeAccel * 0.6
		runAccel = self.runAccel * 0.6
	end

	if kbd.isDown("left") then
		vx = vx - dt * (vx > 0 and brakeAccel or runAccel)
		self.facing = -1
	elseif kbd.isDown("right") then
		vx = vx + dt * (vx < 0 and brakeAccel or runAccel)
		self.facing = 1
	else
		local brake = dt * (vx < 0 and brakeAccel or -runAccel)
		if math.abs(brake) > math.abs(vx) then
			vx = 0
		else
			vx = vx + brake
		end
	end

	if not self.onLadder then
		if kbd.isDown(keyJump) then
			self.jumpHeightMax = 200
		else
			self.jumpHeightMax = 48
		end

		if kbd.isDown("up") then
			if self.jumpCooldown <= 0 then
				if (self.chestl == 9 and self.chestc == 9 and self.chestr == 9) then
					self.onLadder = true
					vx = 0
				elseif self.feetl == 9 and self.feetc == 9 then
					self.x = self.x - dt * 32
					self.onLadder = true
				elseif self.feetc == 9 and self.feetr == 9 then
					self.x = self.x + dt * 32
					self.onLadder = true
				end
			end
		elseif kbd.isDown("down") then
			if self.jumpCooldown <= 0 then
				if (self.tlb == 9 and self.tcb == 9 and self.trb == 9) then
					if not self.isJumping then
						self.onLadder = true
						vx = 0
					end
				elseif self.tlb == 9 and self.tcb == 9 then
					self.x = self.x - dt * 32
				elseif self.tcb == 9 and self.trb == 9 then
					self.x = self.x + dt * 32
				end
			end
		end
	else
		vy = 0
		if kbd.isDown("down") then
			self.y = self.y + dt * self.ladderClimbSpeed
			if self.tcb ~= 9 and self.chestc ~= 9 then
				self.onLadder = false
			end
			self.moving = true
		elseif kbd.isDown("up") then
			self.y = self.y - dt * self.ladderClimbSpeed
			self.moving = true
		else
			self.moving = false
		end
	end


	self.vx, self.vy = vx, vy
end

function smbPlayer:changeVelocityByGravity(dt)
	if (self.tcb ~= 9 and self.feetc ~= 9) then self.onLadder = false end
	if self.onLadder then return end
	local grav = self.gravityAccel
	if self.inWater then grav = 50 end
	self.vy = self.vy + grav * dt
end

function smbPlayer:changeVelocityByCollisionNormal(nx, ny, bounciness)
	if self.onLadder then return end
	bounciness = bounciness or 0
	local vx, vy = self.vx, self.vy

	if (nx < 0 and vx > 0) or (nx > 0 and vx < 0) then
		vx = -vx * bounciness
	end

	if (ny < 0 and vy > 0) or (ny > 0 and vy < 0) then
		vy = -vy * bounciness
	end

	self.vx, self.vy = vx, vy
end

function smbPlayer:move(l,t)
  self.x, self.y = l,t
  self.world:move(self, l, t)
end

function smbPlayer:getCenter()
  return self.x + self.w / 2,
         self.y + self.h / 2
end

function smbPlayer:destroy()
  self.world:remove(self)
end
