local lgr = love.graphics
local win = love.window
local tmr = love.timer
local lfs = love.filesystem
local kbd = love.keyboard
local string = string
local tonumber = tonumber
local collectgarbage = collectgarbage
local sort = table.sort

threeDeeWorld = {}
local drawnTextures = 0
local drawnSprites = 0
local calculatedPoints = 0

function threeDeeWorld:create(v)
	local _i = {}
	setmetatable(_i, {__index = self})
	_i:setup(v or {})
	return _i
end
function threeDeeWorld:setup(v)
	self.width = v.width or screen.drawWidth
	self.height = v.height or screen.drawHeight
	print("Size:", self.width, self.height)

	local curves = { [480] = 28, [720] = 20, [1440] = 17, [1080] = 13, [2160] = 6 }
	self.curveAmount = curves[self.height]
	print(self.curveAmount)

	self.forceCreateCache = true

	-- 480 = 28
	-- 720 = 20
	-- 1080 = 12
	-- 2160 = 6

	self.loaded = false

	self.zoomScale = screen.zoomScale
	self.renderEntireWorld = v.renderEntireWorld or false

	-- self.level = v.level or "battle1"

	-- self:loadLevel(self.level)

	self.skyColorTable = {
		background = {},
		clouds = {},
		fog = {},
		intensity = {}
	}
	local id1 = love.image.newImageData("images/sky/skycolor.png")
	local id2 = love.image.newImageData("images/sky/fogcolor.png")
	local id3 = love.image.newImageData("images/sky/fogintensity.png")
	local id4 = love.image.newImageData("images/sky/cloudcolor.png")
	self.skyColorImageWidth = id2:getWidth()
	local j = math.floor(self.skyColorImageWidth * 0.75)
	for i = 0, id2:getWidth() - 1 do
		local r, g, b, _ = id1:getPixel(i, 0)
		self.skyColorTable.background[j] = { r, g, b }
		local r, g, b, _ = id2:getPixel(i, 0)
		self.skyColorTable.fog[j] = { r, g, b }
		local _, _, _, a = id3:getPixel(i, 0)
		self.skyColorTable.intensity[j] = 1 - math.round(a / 255, 1)
		local r, g, b, _ = id4:getPixel(i, 0)
		self.skyColorTable.clouds[j] = { r, g, b, 200 }
		j = j + 1
		if j >= self.skyColorImageWidth then j = 0 end
	end

	self.skyCanvas = lgr.newCanvas(2048, 256)
	self.skyCanvas:clear(255,0,0)
	self.skyCanvas:setFilter("linear")
	self.cloudQuads = {
		[1] = lgr.newQuad(0, 0, 672, 128, imageLibrary["clouds"]:getWidth(), imageLibrary["clouds"]:getHeight()),
		[2] = lgr.newQuad(672, 0, 352, 128, imageLibrary["clouds"]:getWidth(), imageLibrary["clouds"]:getHeight())
	}

	self.clouds = {
		[1] = { x = 0, speed = 2, quad = 1, width = 672 },
		[2] = { x = 650, speed = 4, quad = 2, width = 352 },
		[3] = { x = 800, speed = 3, quad = 1, width = 672 },
	}

	self.timeSpeed = v.timeSpeed or 0.02
	self.timeOfDay = v.timeOfDay or 6
	self.isDaytime = v.isDaytime or true
	self.dayCount = 1

	self.curvature = 0

	self.timer = 0

	collectgarbage("setpause", 500)
	collectgarbage("setstepmul", 500)
	-- self.collector = cron.every(5, function() collectgarbage("step") end)
	fogShader.send(fogShader, "tint", {0,0,0,0})
end


-- 88   88 88""Yb 8888b.     db    888888 888888
-- 88   88 88__dP  8I  Yb   dPYb     88   88__
-- Y8   8P 88"""   8I  dY  dP__Yb    88   88""
-- `YbodP' 88     8888Y"  dP""""Yb   88   888888
function threeDeeWorld:update(dt)
	self.timer = self.timer + dt

	self.lightingCanvas:clear(255,255,255,255)

	if not self.disableEnvironment then
		local smod = 1
		if kbd.isDown("rshift") then smod = 200 end
		self.timeOfDay = self.timeOfDay + (dt * self.timeSpeed) * smod
		self.isDaytime = (self.timeOfDay < 12)
		if self.timeOfDay > 24 then
			self.timeOfDay = self.timeOfDay - 24
			self.dayCount = self.dayCount + 1
		end
		self.timeOfDaySeg = math.floor((self.timeOfDay / 24) * self.skyColorImageWidth)
	else
		self.timeOfDaySeg = 0
	end

	drawnTextures = 0
	drawnSprites = 0
	calculatedPoints = 0

	local cam = self.camera
	if cam.focusOnPosition then
		cam.locX = math.floor(cam.x + math.sin(-math.rad(cam.tiltz)) * self.focal)
		cam.locZ = math.floor(cam.z + math.cos(-math.rad(cam.tiltz)) * self.focal)
	else
		cam.locX = cam.x
		cam.locZ = cam.z
	end

	self.worldPoints = {}
	self.worldObjects = {}
	self.worldSprites = {}

	self:calculateCameraRotation(cam)

	if performanceGraph.enabled then
		self:debugDrawHotzones()
	end

	self.chunksDrawnThisFrame = {}

	for i, e in pairs(self.entity) do
		e:update(dt)
	end

	if performanceGraph.enabled then
		for i, e in pairs(game.entity) do
			if e.map == self.loadedMap then
				if e.moving then
					local n = #self.worldObjects+1
					local cx1, cy1, dist, _, onScreen = self:point3dto2d(e.x, e.y, e.z, cam)
					local cx2, cy2, _, scale, onScreen = self:point3dto2d(e.target.x, 0, e.target.z, cam)
					if onScreen then
						self.worldObjects[n] = { x1 = cx1, y1 = cy1, x2 = cx2, y2 = cy2, origX = 0, origY = 0, origZ = 0, color = {255,0,0},
								dist = dist, thickness = 2 * scale, kind = "line", layer = 10 }
					end
				elseif e.waiting then
					local n = #self.worldObjects+1
					local cx1, cy1, dist, scale, onScreen = self:point3dto2d(e.x, e.y + 10, e.z, cam)
					local s = e.waitTime / e.totalWaitTime
					if onScreen then
						self.worldObjects[n] = { points = {cx1 - 50, cy1, cx1 + 50, cy1, cx1 + 50, cy1 + 10, cx1 - 50, cy1 + 10},
								color = {255,255,255,100}, dist = dist, kind = "poly", layer = 10 }
						self.worldObjects[n+1] = { points = {cx1 - 50, cy1, cx1 - 50 + (100 * s), cy1, cx1 - 50 + (100 * s), cy1 + 10, cx1 - 50, cy1 + 10},
								color = {0,255,0}, dist = dist-1, kind = "poly", layer = 10 }
					end
				end
			end
		end
	end

	local xlb, xub
	local zlb, zub
	if self.renderEntireWorld then
		xlb, xub = math.floor((cam.x-self.fieldOfView) / self.chunkSize), math.floor((cam.x+self.fieldOfView) / self.chunkSize)
		zlb, zub = math.floor((cam.z-self.fieldOfView) / self.chunkSize), math.floor((cam.z+self.fieldOfView) / self.chunkSize)
	else
		xlb, xub = math.floor((cam.x-self.fieldOfView) / self.chunkSize), math.floor((cam.x+self.fieldOfView) / self.chunkSize)
		zlb, zub = math.floor((cam.z-self.focal) / self.chunkSize), math.floor((cam.z+self.maxDistance) / self.chunkSize)
	end
	for z = zub, zlb, -1 do
		-- if z == zlb or z == zlb + 1 then
		-- 	xlb = xlb + 1
		-- 	xub = xub - 1
		-- end
		for x = xlb, xub do
			local chunkName = "c"..x.."_"..z
			if self.chunkList[chunkName] then
				for i, p in pairs(self.chunkList[chunkName].points) do
					local pp = self.points[self.chunkList[chunkName].points[i]]
					local cx, cy, dist, scale, onScreen = self:point3dto2d(pp.x, pp.y, pp.z, cam)
					calculatedPoints = calculatedPoints + 1
					self.worldPoints[self.chunkList[chunkName].points[i]] = { cx = cx, cy = cy, dist = dist, scale = scale, onScreen = onScreen }
				end


				local _timeForPolyGen = tmr.getTime()
				for i, p in pairs(self.chunkList[chunkName].polys) do
					local pp = self.polys[self.chunkList[chunkName].polys[i]]

					if self:onScreen(self.worldPoints[pp.points[1]], self.worldPoints[pp.points[2]], self.worldPoints[pp.points[3]], self.worldPoints[pp.points[4]], self.width * self.zoomScale, self.height * self.zoomScale) then
						local n = #self.worldObjects+1
						local dist
						if pp.sortPoint then
							dist = self.worldPoints[pp.points[pp.sortPoint]].dist
						else
							dist = math.min(self.worldPoints[pp.points[1]].dist, self.worldPoints[pp.points[2]].dist,
											self.worldPoints[pp.points[3]].dist, self.worldPoints[pp.points[4]].dist)
						end
						local dim = self.fogIntensity
						if dist > self.fogDistance then
							dim = math.min(1 - ((dist - self.fogDistance) / (self.maxDistance - self.fogDistance)), self.fogIntensity)
						end

						if dim > 0 then
							self.worldObjects[n] = { points = { self.worldPoints[pp.points[1]].cx, self.worldPoints[pp.points[1]].cy,
																self.worldPoints[pp.points[2]].cx, self.worldPoints[pp.points[2]].cy,
																self.worldPoints[pp.points[3]].cx, self.worldPoints[pp.points[3]].cy,
																self.worldPoints[pp.points[4]].cx, self.worldPoints[pp.points[4]].cy, },
													color = pp.color, dist = dist, kind = "poly", image = pp.image, layer = pp.layer,
													chunk = pp.chunk, flip = pp.flip, dim = math.max(dim - (pp.dim or 0), 0) }
						end
					end
				end
				performanceGraph:addToWorldPolyGenTime(tmr.getTime() - _timeForPolyGen)

				for i, s in pairs(self.chunkList[chunkName].sprites) do
					local sp = self.sprites[s]
					local sx = self.worldPoints[sp.point].cx
					local sy = self.worldPoints[sp.point].cy

					local scale = self.worldPoints[sp.point].scale * sp.scale
					if sp.absoluteScale then
						scale = sp.scale
					end

					local dist = self.worldPoints[sp.point].dist
					local dim = self.fogIntensity
					if dist > self.fogDistance then
						dim = math.min(1 - ((dist - self.fogDistance) / (self.maxDistance - self.fogDistance)), self.fogIntensity)
					end
					if dist < self.maxDistance then
						if (not self.threeDeeSprites or sp.forceSprite) and not sp.forceThreeDee then
							local n = #self.worldObjects+1
							self.worldObjects[n] = { x = sx, y = sy, origX = self.points[sp.point].x, origY = self.points[sp.point].y, origZ = self.points[sp.point].z,
														color = sp.color, image = sp.image, angle = sp.angle, dist = dist, dim = dim,
														scale = scale, kind = "sprite", layer = sp.layer, offsetX = sp.offsetX, offsetY = sp.offsetY }
						else
							self:generate3DSprite(sp, cam, chunkName)
						end
					end
				end
			end
			self.chunksDrawnThisFrame[#self.chunksDrawnThisFrame+1] = chunkName
		end
	end

	for i, p in pairs(self.globalPoints) do
		local pp = self.points[p]
		-- print(chunkName, i, pp.x, pp.y, pp.z, self.chunkList[chunkName].points[i])
		local cx, cy, dist, scale, onScreen = self:point3dto2d(pp.x, pp.y, pp.z, cam)
		calculatedPoints = calculatedPoints + 1
		self.worldPoints[p] = { cx = cx, cy = cy, dist = dist, scale = scale, onScreen = onScreen }
	end

	for i, s in pairs(self.globalSprites) do
		local sp = self.sprites[s]
		local sx = self.worldPoints[sp.point].cx
		local sy = self.worldPoints[sp.point].cy
		if sx and sy then
			if sx > -32 and sy > -32 and sx < self.width * self.zoomScale + 32 and sy < self.height * self.zoomScale + 32 and self.worldPoints[sp.point].dist < self.maxDistance then
				local n = #self.worldObjects+1
				local scale = self.worldPoints[sp.point].scale * sp.scale
				if sp.absoluteScale then
					scale = sp.scale
				end

				if sp.illumination > 0 then
					lgr.setCanvas(self.lightingCanvas)
					lgr.setBlendMode("subtractive")
					lgr.setColor(255,255,255)
					lgr.circle("fill", self.worldPoints[sp.point].cx, self.worldPoints[sp.point].cy - (sp.scale * imageLibrary[sp.image]:getHeight()), 500)
					lgr.setBlendMode("alpha")
					lgr.setCanvas()
				end

				local dist = self.worldPoints[sp.point].dist
				local dim = self.fogIntensity
				if dist > self.fogDistance then
					dim = math.min(1 - ((dist - self.fogDistance) / (self.maxDistance - self.fogDistance)), self.fogIntensity)
				end
				if (not self.threeDeeSprites or self.camera.tiltz ~= 0 or sp.forceSprite) and not sp.forceThreeDee then
					self.worldObjects[n] = { x = self.worldPoints[sp.point].cx, y = self.worldPoints[sp.point].cy,
												origX = self.points[sp.point].x, origY = self.points[sp.point].y, origZ = self.points[sp.point].z,
												color = sp.color, image = sp.image, angle = sp.angle, dist = dist, dim = dim,
												scale = scale, kind = "sprite", layer = sp.layer, offsetX = sp.offsetX, offsetY = sp.offsetY }
				else
					self:generate3DSprite(sp, cam)
				end
			end
		end
	end

	-- local n = #self.worldObjects+1
	-- self.worldObjects[n] = { x1 = 0, y1 = 0, x2 = 400, y2 = 500, origX = 0, origY = 0, origZ = 0, color = {255,0,0}, dist = 10, thickness = 3, kind = "line", layer = 10 }

	sort(self.worldObjects,
		function(a, b)
			if a.layer ~= b.layer then return a.layer < b.layer end
			return a.dist > b.dist
		end
	)

	if not self.disableEnvironment then
		self.skyColor = self.skyColorTable.background[self.timeOfDaySeg]
		self.fogColor = self.skyColorTable.fog[self.timeOfDaySeg]
		self.fogIntensity = self.skyColorTable.intensity[self.timeOfDaySeg]
		self.cloudColor = self.skyColorTable.clouds[self.timeOfDaySeg]
		if cam.tilty < 45 then
			lgr.setCanvas(self.skyCanvas)
			self.skyCanvas:clear(self.skyColor)
			lgr.setColor(255,255,255)
			local offset = 0
			if not self.isDaytime then
				offset = 12
			end
			local smx = (1 - ((self.timeOfDay - offset) / 12)) * 1124 - 50
			local smy = 200 - (math.sin((1 - ((self.timeOfDay - offset) / 12)) * math.pi) * 80)
			if not self.isDaytime then
				lgr.draw(imageLibrary["moon"], smx, smy, 0, 0.5, 0.5, 128, 128)
			else
				lgr.draw(imageLibrary["sun"], smx, smy, 0, 0.5, 0.5, 128, 128)
			end
			lgr.setColor(self.cloudColor)
			for i = 1, #self.clouds do
				local c = self.clouds[i]
				c.x = c.x - dt * c.speed
				if c.x < -c.width then c.x = 1024 end
				lgr.draw(imageLibrary["clouds"], self.cloudQuads[c.quad], c.x, 128)
			end
			lgr.setColor(self.fogColor)
			lgr.draw(imageLibrary["gradient"], 0, 256, 0, 2048, -1.5, 0, 0)
			-- lgr.setColor(0,0,0,100)
			-- lgr.draw(imageLibrary["gradient"], 0, 0, 0, 2048, 1.5, 0, 0)
			lgr.setCanvas()
		end
	else
		self.skyColor = self.backgroundColor or {0,0,0}
		self.fogColor = {0,0,0}
		self.fogIntensity = self.ambientDim
		-- self.cloudColor = {255,255,255}
	end
end


-- 8888b.  88""Yb    db    Yb        dP
--  8I  Yb 88__dP   dPYb    Yb  db  dP
--  8I  dY 88"Yb   dP__Yb    YbdPYbdP
-- 8888Y"  88  Yb dP""""Yb    YP  YP
function threeDeeWorld:draw()
	local cam = self.camera

	lgr.setColor(self.skyColor)
	lgr.rectangle("fill", 0, 0, self.width, self.height)

	if cam.tilty < 45 and not self.disableEnvironment then
		local skyScale = screen.zoomScale / 3.1
		local cx, cy, dist, scale, onScreen = self:point3dto2d(0, 0, cam.z + 3000, cam)
		if cx and cy then
			cx = self.width / 2
			lgr.push()
			lgr.translate(cx, cy)
			lgr.scale((screen.pixelScale * 2) / screen.pixelScale)
			lgr.setColor(self.fogColor)
			lgr.rectangle("fill", -1000, 0, 2000, (30 * skyScale))
			lgr.setColor(255,255,255)
			lgr.draw(self.skyCanvas, 0, 0, 0, 0.65 * skyScale, 0.65 * skyScale, 512, 256)
			lgr.draw(self.skyCanvas, 0, 30 * skyScale, 0, 0.65 * skyScale, -(0.65 * skyScale), 512, 256)
			lgr.pop()
		end
	end


	for i, p in pairs(self.worldObjects) do
		local dim = 1--math.min(math.max(1 - (p.dist / 2500 * 2), 0), 1)
		if p.kind == "poly" then
			local bf = self:backface(p.points[1], p.points[2], p.points[3], p.points[4], p.points[5], p.points[6], p.points[7], p.points[8])
			if bf then
				if self.drawTextures then
					if p.color then
						local c = p.color
						local f = self.fogColor
						local dim = p.dim or 1
						local diff = {f[1] - c[1], f[2] - c[2], f[3] - c[3]}

						lgr.setColor(
							c[1] + diff[1] * (1-dim),
							c[2] + diff[2] * (1-dim),
							c[3] + diff[3] * (1-dim),
							c[4] or 255
						)
						lgr.polygon("fill", p.points)
						if p.border then
							lgr.setColor(0,0,0)
							lgr.polygon("line", p.points)
						end
					end

					if p.image then
						local tex
						if type(p.image) == "table" then
							tex = {}
							for k = 1, #p.image do
								tex[k] = imageLibrary[p.image[k]]
							end
						else
							tex = imageLibrary[p.image]
						end
						local p1, p2, p3, p4
						if p.flip then
							p2 = {(p.points[5]), (p.points[6])}
							p1 = {(p.points[7]), (p.points[8])}
							p4 = {(p.points[1]), (p.points[2])}
							p3 = {(p.points[3]), (p.points[4])}
						else
							p1 = {(p.points[5]), (p.points[6])}
							p2 = {(p.points[7]), (p.points[8])}
							p3 = {(p.points[1]), (p.points[2])}
							p4 = {(p.points[3]), (p.points[4])}
						end

						local dim = p.dim or 1
						if dim > 0 then
							local h = tonumber(self.height)
							local w = tonumber(win.getWidth())
							local t = tmr.getTime()
							lgr.setColor(0,0,0,0)
							if type(tex) == "table" then
								for j = 1, #tex do
									texture.quad(tex[j],p1,p2,p3,p4,w,h,{self.fogColor[1]/255,self.fogColor[2]/255,self.fogColor[3]/255,1-dim},self.lightingCanvas)
									drawnTextures = drawnTextures + 1
								end
							else
								texture.quad(tex,p1,p2,p3,p4,w,h,{self.fogColor[1]/255,self.fogColor[2]/255,self.fogColor[3]/255,1-dim},self.lightingCanvas)
								drawnTextures = drawnTextures + 1
							end
							if performanceGraph then performanceGraph:addToTextureTime(tmr.getTime() - t) end
						end
					end
				else
					local c, o
					if p.chunk then
						c = self.chunkList[p.chunk].dbgColor
					else
						c = {250,140,0}
					end
					if p.image == "water" then o = 0.6 else o = 1 end
					lgr.setColor(c[1] * dim, c[2] * dim, c[3] * dim, 255 * o)
					lgr.polygon("fill", p.points)
					lgr.setColor(c[1] * (dim * 0.25), c[2] * (dim * 0.25), c[3] * (dim * 0.25))
					lgr.polygon("line", p.points)
					lgr.setColor(0,0,0)
					lgr.print(tostring(p.layer), p.points[1], p.points[2])
				end
				if p.text then
					lgr.push()
					lgr.translate((p.points[1]), (p.points[2]))
					lgr.scale(2 * screen.debugTextScale)
					lgr.setFont(font.tiny)
					lgr.setColor(0,0,0)
					lgr.print(p.text, 1, 0)
					lgr.setColor(255,255,255)
					lgr.print(p.text, 0, 0)
					lgr.pop()
				end
			end
		elseif p.kind == "sprite" then
			if p.x and p.y then
				local op = 255
				if not self.drawTextures then op = 100 end
				local rot = math.rad(cam.tiltx)
				-- math.cos(x + game.timer * 1.2) * math.sin(z + game.timer * 5.5) * 4
				local scaleX = p.scale
				local scaleY = p.scale
				local shear = 0
				-- local shear2 = 0
				if p.image == "tree" or p.image == "tree1" or p.image == "tree2" or p.image == "tree1b" or p.image == "tree2b" or p.image == "tree3b" or p.image == "tree4b" then
					shear = math.cos(p.origX + game.timer * 5 + (p.origZ * 5)) / 50
					-- shear2 = math.sin(p.origX + game.timer * 10) / 500--easing.inOutBounce(game.timer % 1, 0, 1, 1)
				end
				if p.image == "flower1" or p.image == "flower2" or p.image == "flower3" or p.image == "flower4" or p.image == "grass1" or p.image == "grass2" or p.image == "grass3" then
					shear = math.cos(p.origX + game.timer * 2 + (p.origZ * 5)) / 10
				end
				if p.image == "fountain" then
					scaleX = math.cos(game.timer * 50)
					-- shear2 = math.sin(p.origX + game.timer * 10) / 500--easing.inOutBounce(game.timer % 1, 0, 1, 1)
				end
				local dim = p.dim or 1
				love.graphics.setShader(fogShader)
				fogShader.send(fogShader, "tint", {self.fogColor[1]/255,self.fogColor[2]/255,self.fogColor[3]/255,1-dim})
				lgr.setColor(255,255,255,op)
				lgr.draw(imageLibrary[p.image], math.floor(p.x), math.floor(p.y), rot, scaleX, scaleY, p.offsetX, p.offsetY, shear, 0)
				love.graphics.setShader()

				drawnSprites = drawnSprites + 1
			end
		elseif p.kind == "line" then
			local oldThickness = lgr.getLineWidth()
			lgr.setLineWidth(p.thickness)
			lgr.setColor(p.color)
			lgr.line(p.x1, p.y1, p.x2, p.y2)
			lgr.setLineWidth(oldThickness)
		end
	end

	if not self.drawTextures then
		for i, p in pairs(self.worldPoints) do
			if p.cx and p.cy then
				lgr.setColor(255,255,0)
				lgr.circle("fill", p.cx, p.cy, 1 * screen.pixelScale * p.scale, 4)
				-- lgr.rectangle("fill", p.cx-(1*screen.pixelScale*p.scale), p.cy-(1*screen.pixelScale*p.scale), (2*screen.pixelScale*p.scale), (2*screen.pixelScale*p.scale))
			end
		end
	end

	self:drawDebugInfo()

	if kbd.isDown("c") then
		local s = screen.drawHeight / self.gridCanvas:getHeight()
		lgr.push()
		lgr.translate(0, screen.drawHeight)

		lgr.push()
		lgr.scale(s, -s)
		lgr.setColor(255,255,255)
		lgr.draw(self.gridCanvas, 0, 0, 0, 1, 1)

		lgr.scale(0.25)
		local l = self.colWorld:getRect(game.worldMapPos)
		lgr.setColor(0,255,0)
		lgr.rectangle("fill", l.l, l.t, l.w, l.h)

		for i, e in pairs(game.entity) do
			if e.map == game_worldmap.world.loadedMap or e.alwaysUpdate then
				lgr.push()
				lgr.translate(math.floor(e.x), math.floor(e.z))
				lgr.setColor(255,255,0)
				lgr.rectangle("fill", -e.w/2, -e.h/2, e.w, e.h)
				-- lgr.circle("fill", 0, 0, 20)
				lgr.setFont(font.dialog)
				lgr.scale(4, -4)
				lgr.setColor(255,255,255)
				lgr.printf(e.nameID, -50, 5, 100, "center")
				lgr.pop()
			end
		end
		lgr.pop()

		lgr.pop()
	end

	self:drawMiniMap()

	self:drawWelcomeSign()

	drawnTextures = 0
	drawnSprites = 0
end

-- 88  dP 888888 Yb  dP 88""Yb 88""Yb 888888 .dP"Y8 .dP"Y8 888888 8888b.
-- 88odP  88__    YbdP  88__dP 88__dP 88__   `Ybo." `Ybo." 88__    8I  Yb
-- 88"Yb  88""     8P   88"""  88"Yb  88""   o.`Y8b o.`Y8b 88""    8I  dY
-- 88  Yb 888888  dP    88     88  Yb 888888 8bodP' 8bodP' 888888 8888Y"
function threeDeeWorld:keypressed(k, r)
	if k == "f3" then
		self.drawTextures = not self.drawTextures
	elseif k == "f2" then
		self.threeDeeSprites = not self.threeDeeSprites
	elseif k == "f5" then
		self.entity["door_store"]:open()
	end
end


-- 8b    d8  dP"Yb  88   88 .dP"Y8 888888 88""Yb 88""Yb 888888 .dP"Y8 .dP"Y8 888888 8888b.
-- 88b  d88 dP   Yb 88   88 `Ybo." 88__   88__dP 88__dP 88__   `Ybo." `Ybo." 88__    8I  Yb
-- 88YbdP88 Yb   dP Y8   8P o.`Y8b 88""   88"""  88"Yb  88""   o.`Y8b o.`Y8b 88""    8I  dY
-- 88 YY 88  YbodP  `YbodP' 8bodP' 888888 88     88  Yb 888888 8bodP' 8bodP' 888888 8888Y"
function threeDeeWorld:mousepressed(x, y, button)
	-- print("3D World", x, y, button)
	if button == "wu" then
		if kbd.isDown("lalt", "ralt") then
			self.fieldOfView = math.min(self.fieldOfView + 64, 2048)
		else
			self.maxDistance = math.min(self.maxDistance + 64, 15000)
		end
	elseif button == "wd" then
		if kbd.isDown("lalt", "ralt") then
			self.fieldOfView = math.max(self.fieldOfView - 64, 512)
		else
			self.maxDistance = math.max(self.maxDistance - 64, 512)
		end
	end
end


--  dP""b8 888888 88b 88 888888 88""Yb    db    888888 888888     88888 8888b.      .dP"Y8 88""Yb 88""Yb 88 888888 888888
-- dP   `" 88__   88Yb88 88__   88__dP   dPYb     88   88__         .dP  8I  Yb     `Ybo." 88__dP 88__dP 88   88   88__
-- Yb  "88 88""   88 Y88 88""   88"Yb   dP__Yb    88   88""       o `Yb  8I  dY     o.`Y8b 88"""  88"Yb  88   88   88""
--  YboodP 888888 88  Y8 888888 88  Yb dP""""Yb   88   888888     YbodP 8888Y"      8bodP' 88     88  Yb 88   88   888888
function threeDeeWorld:generate3DSprite(sp, cam, chunkName)
	local _timeForSpriteGen = tmr.getTime()
	local scaleX = sp.scale
	local scaleY = sp.scale
	local shear = 0
	if string.sub(sp.image, 0, 4) == "tree" then
	-- if sp.image == "tree_top" or sp.image == "tree1" or sp.image == "tree2" or sp.image == "tree1b" or sp.image == "tree2b" or sp.image == "tree3b" or sp.image == "tree4b" then
		shear = math.cos(self.points[sp.point].x + game.timer * 3 + (self.points[sp.point].z * 3))
	end
	if sp.image == "tree1" or sp.image == "tree2" or sp.image == "tree1b" or sp.image == "tree2b" or sp.image == "tree3b" or sp.image == "tree4b" then
		shear = math.cos(self.points[sp.point].x + game.timer * 5 + (self.points[sp.point].z * 5))
		-- shear2 = math.sin(self.points[sp.point].x + game.timer * 10) / 500--easing.inOutBounce(game.timer % 1, 0, 1, 1)
	end
	if sp.image == "flower1" or sp.image == "flower2" or sp.image == "flower3" or sp.image == "flower4" or sp.image == "flower5" or sp.image == "grass1" or sp.image == "grass2" or sp.image == "grass3" then
		shear = math.cos(self.points[sp.point].x + game.timer * 2 + (self.points[sp.point].z * 5))
	end
	if sp.image == "fountain" then
		scaleX = math.cos(game.timer * 50)
		-- shear2 = math.sin(self.points[sp.point].x + game.timer * 10) / 500--easing.inOutBounce(game.timer % 1, 0, 1, 1)
	end

	local image = sp.image

	local iw = imageLibrary[sp.image]:getWidth()
	local ih = imageLibrary[sp.image]:getHeight()
	local baseX, baseY, baseZ = self.points[sp.point].x, self.points[sp.point].y, self.points[sp.point].z
	if string.sub(sp.image, 1, 6) == "player" then
		baseY = baseY - self.flightElevation
		if self.flightMode and self.flightElevation > 0 then
			baseY = baseY + math.math.sin(game.timer * 4) * 3
		end
	end
	local imgWidth = iw * sp.scale
	local imgHeight = ih * sp.scale * 1.1
	local top, bottom, left, right, far, near
	local newPoints
	if string.sub(sp.image, 1, 6) == "flower" then
		top =		baseY - 10
		bottom =	baseY - 2
		left =		baseX - imgWidth / 2
		right =		baseX + imgWidth / 2
		far =		baseZ + imgHeight / 2
		near =		baseZ - imgHeight / 2
		image =		image .. "top"

		newPoints = {
			{right + shear * 2, bottom + shear, near, math.random(#self.points+1,999999999)},
			{left + shear * 2, bottom + -shear, near, math.random(#self.points+1,999999999)},
			{left + shear * 2, top + -shear, far, math.random(#self.points+1,999999999)},
			{right + shear * 2, top + shear, far, math.random(#self.points+1,999999999)},
		}
	elseif string.sub(sp.image, 1, 8) == "fountain" then
		top =		baseY + -imgHeight
		bottom =	baseY
		left =		baseX - (imgWidth / 2) * scaleX
		right =		baseX + (imgWidth / 2) * scaleX
		far =		baseZ
		near =		baseZ

		newPoints = {
			{right, bottom, near, math.random(#self.points+1,999999999)},
			{left, bottom, near, math.random(#self.points+1,999999999)},
			{left + shear, top + -shear, near, math.random(#self.points+1,999999999)},
			{right + shear, top + shear, near, math.random(#self.points+1,999999999)},
		}
	elseif string.sub(sp.image, 1, 6) == "shadow" then
		top =		baseY
		bottom =	baseY
		left =		baseX - imgWidth / 2
		right =		baseX + imgWidth / 2
		far =		baseZ + imgHeight / 2 * 0.7
		near =		baseZ - imgHeight / 2 * 0.7

		newPoints = {
			{right, bottom, near, math.random(#self.points+1,999999999)},
			{left, bottom, near, math.random(#self.points+1,999999999)},
			{left, top, far, math.random(#self.points+1,999999999)},
			{right, top, far, math.random(#self.points+1,999999999)},
		}
	elseif string.sub(sp.image, 1, 8) == "tree_top" then
		top =		baseY + -(imgHeight*0.9)
		bottom =	baseY
		left =		baseX - imgWidth / 2
		right =		baseX + imgWidth / 2
		far =		baseZ + 32
		near =		baseZ

		newPoints = {
			{right, bottom, near, math.random(#self.points+1,999999999)},
			{left, bottom, near, math.random(#self.points+1,999999999)},
			{left + shear, top + -shear, far, math.random(#self.points+1,999999999)},
			{right + shear, top + shear, far, math.random(#self.points+1,999999999)},
		}
	else
		top =		baseY + -imgHeight
		bottom =	baseY
		left =		baseX - imgWidth / 2
		right =		baseX + imgWidth / 2
		far =		baseZ
		near =		baseZ

		newPoints = {
			{right, bottom, near, math.random(#self.points+1,999999999)},
			{left, bottom, near, math.random(#self.points+1,999999999)},
			{left + shear, top + -shear, near, math.random(#self.points+1,999999999)},
			{right + shear, top + shear, near, math.random(#self.points+1,999999999)},
		}
	end


	for j = 1, #newPoints do
		local n = newPoints[j][4]
		local cx, cy, dist, scale, onScreen = self:point3dto2d(newPoints[j][1], newPoints[j][2], newPoints[j][3], cam)
		calculatedPoints = calculatedPoints + 1
		self.worldPoints[n] = { cx = cx, cy = cy, dist = dist, scale = scale, onScreen = onScreen }
	end

	local pts = {
		self.worldPoints[newPoints[1][4]].cx, self.worldPoints[newPoints[1][4]].cy,
		self.worldPoints[newPoints[2][4]].cx, self.worldPoints[newPoints[2][4]].cy,
		self.worldPoints[newPoints[3][4]].cx, self.worldPoints[newPoints[3][4]].cy,
		self.worldPoints[newPoints[4][4]].cx, self.worldPoints[newPoints[4][4]].cy,
	}

	local n = #self.worldObjects+1
	local dist = self.worldPoints[newPoints[1][4]].dist
	local l = sp.layer
	local dim = self.fogIntensity
	if dist > self.fogDistance then
		dim = math.min(1 - ((dist - self.fogDistance) / (self.maxDistance - self.fogDistance)), self.fogIntensity)
	end
	self.worldObjects[n] = { points = pts, dist = dist, kind = "poly",
								image = image, layer = l, chunk = chunkName, flip = sp.flip, dim = dim }
	-- print(n)

	-- print(i, sp.image, imgWidth, imgHeight, top, bottom, left, right)
	performanceGraph:addToWorldSpriteGenTime(tmr.getTime() - _timeForSpriteGen)
end


--  dP""b8 888888 88b 88 888888 88""Yb    db    888888 888888     88""Yb  dP"Yb  88 88b 88 888888      dP""b8    db     dP""b8 88  88 888888
-- dP   `" 88__   88Yb88 88__   88__dP   dPYb     88   88__       88__dP dP   Yb 88 88Yb88   88       dP   `"   dPYb   dP   `" 88  88 88__
-- Yb  "88 88""   88 Y88 88""   88"Yb   dP__Yb    88   88""       88"""  Yb   dP 88 88 Y88   88       Yb       dP__Yb  Yb      888888 88""
--  YboodP 888888 88  Y8 888888 88  Yb dP""""Yb   88   888888     88      YbodP  88 88  Y8   88        YboodP dP""""Yb  YboodP 88  88 888888
function threeDeeWorld:calcPointCache()
	for i = 1, #self.points do
		local n = tostring("p" .. self.points[i].x .. "_" .. self.points[i].y .. "_" .. self.points[i].z)
		self.pointCache[n] = i
	end
end


--    db    8888b.  8888b.      88""Yb  dP"Yb  88     Yb  dP  dP""b8  dP"Yb  88b 88
--   dPYb    8I  Yb  8I  Yb     88__dP dP   Yb 88      YbdP  dP   `" dP   Yb 88Yb88
--  dP__Yb   8I  dY  8I  dY     88"""  Yb   dP 88  .o   8P   Yb  "88 Yb   dP 88 Y88
-- dP""""Yb 8888Y"  8888Y"      88      YbodP  88ood8  dP     YboodP  YbodP  88  Y8
function threeDeeWorld:addPolygon(p, c, image, layer, flip, sortPoint, dim)
	local layer = layer or 0
	local flip = flip or false
	local sortPoint = sortPoint or nil
	local c = c or nil
	local image = image or nil
	local sp = {}
	sp[1] = "p" .. p[1][1] .. "_" .. p[1][2] .. "_" .. p[1][3]
	sp[2] = "p" .. p[2][1] .. "_" .. p[2][2] .. "_" .. p[2][3]
	sp[3] = "p" .. p[3][1] .. "_" .. p[3][2] .. "_" .. p[3][3]
	sp[4] = "p" .. p[4][1] .. "_" .. p[4][2] .. "_" .. p[4][3]
	local pts = {}
	for i = 1, 4 do
		if not self.pointCache[sp[i]] then
			local n = #self.points+1
			self.points[n] = {
				x = p[i][1],
				y = p[i][2],
				z = p[i][3],
			}
			-- if self.worldWidth < p[i][1] and p[i][2] == 0 then
			-- 	self.worldWidth = p[i][1]
			-- end
			-- if self.worldDepth < p[i][3] and p[i][2] == 0 then
			-- 	self.worldDepth = p[i][3]
			-- end
			pts[i] = n
			local s = tostring("p" .. p[i][1] .. "_" .. p[i][2] .. "_" .. p[i][3])
			self.pointCache[s] = n
		else
			pts[i] = self.pointCache[sp[i]]
		end
	end
	local n = #self.polys+1
	self.polys[n] = { points = pts, image = image, layer = layer, chunks = {}, flip = flip, sortPoint = sortPoint }
	if c then
		self.polys[n].color = c
	end
	if dim then
		self.polys[n].dim = dim
	end
end

--    db    8888b.  8888b.      .dP"Y8 88""Yb 88""Yb 88 888888 888888
--   dPYb    8I  Yb  8I  Yb     `Ybo." 88__dP 88__dP 88   88   88__
--  dP__Yb   8I  dY  8I  dY     o.`Y8b 88"""  88"Yb  88   88   88""
-- dP""""Yb 8888Y"  8888Y"      8bodP' 88     88  Yb 88   88   888888
function threeDeeWorld:addSprite(v)
	-- name, x, y, z, image, scale, layer, angle
	local v = v or {}
	local layer = v.layer or 0
	local angle = v.angle or 0
	local scale = v.scale or screen.overallScale
	local image = v.image or "default"
	local name = v.name
	local offsetX = v.offsetX or imageLibrary[image]:getWidth() / 2
	local offsetY = v.offsetY or imageLibrary[image]:getHeight()
	local forceSprite = v.forceSprite or nil
	local forceThreeDee = v.forceThreeDee or nil

	local p = "p" .. v.x .. "_" .. v.y .. "_" .. v.z
	-- print("Adding sprite " .. name)

	local n = #self.points+1
	self.points[n] = {
		x = v.x,
		y = v.y,
		z = v.z,
	}

	self.sprites[name] = {
		x = v.x,
		z = v.y,
		w = 32,
		h = 32,
		worldSprite = v.worldSprite or false,
		point = n,
		image = image,
		angle = angle,
		color = {math.random(100,200),math.random(100,200),math.random(100,200)},
		layer = layer,
		scale = scale,
		offsetX = offsetX,
		offsetY = offsetY,
		absoluteScale = v.absoluteScale or false,
		illumination = v.illumination or 0
	}
	if forceSprite then
		self.sprites[name].forceSprite = true
	elseif forceThreeDee then
		self.sprites[name].forceThreeDee = true
	end

	self.colWorld:add(self.sprites[name], self.sprites[name].x, self.sprites[name].z, self.sprites[name].w, self.sprites[name].h)
end

function threeDeeWorld:removeSprite(name)
	self.sprites[name] = nil
end

-- .dP"Y8 888888 888888     .dP"Y8 88""Yb 88""Yb 88 888888 888888     88""Yb  dP"Yb  .dP"Y8 88 888888 88  dP"Yb  88b 88
-- `Ybo." 88__     88       `Ybo." 88__dP 88__dP 88   88   88__       88__dP dP   Yb `Ybo." 88   88   88 dP   Yb 88Yb88
-- o.`Y8b 88""     88       o.`Y8b 88"""  88"Yb  88   88   88""       88"""  Yb   dP o.`Y8b 88   88   88 Yb   dP 88 Y88
-- 8bodP' 888888   88       8bodP' 88     88  Yb 88   88   888888     88      YbodP  8bodP' 88   88   88  YbodP  88  Y8
function threeDeeWorld:setSpritePos(name, x, y, z, l)
	if not self.sprites[name] then return false end
	self.points[self.sprites[name].point].x = x or 0
	self.points[self.sprites[name].point].y = y or 0
	self.points[self.sprites[name].point].z = z or 0
	self.sprites[name].layer = l or self.sprites[name].layer
end

--  dP""b8    db    8b    d8 888888 88""Yb    db
-- dP   `"   dPYb   88b  d88 88__   88__dP   dPYb
-- Yb       dP__Yb  88YbdP88 88""   88"Yb   dP__Yb
--  YboodP dP""""Yb 88 YY 88 888888 88  Yb dP""""Yb
function threeDeeWorld:setupCamera(v)
	local x = v.x or 0
	local y = v.y or 0
	local z = v.z or 0
	local tiltx = v.tiltx or 0
	local tilty = v.tilty or 0
	local tiltz = v.tiltz or 0
	self.camera = {x=x,y=y,z=z, tiltx=tiltx,tilty=tilty,tiltz=tiltz, ptx=-1,pty=-1,ptz=-1, locX=x,locZ=z, focusOnPosition = false}
end

function threeDeeWorld:calculateCameraRotation(cam)
	if cam.pty ~= cam.tilty then
		cam.sx = math.sin(math.rad(cam.tilty))
		cam.cx = math.cos(math.rad(cam.tilty))
	end
	if cam.ptz ~= cam.tiltz then
		cam.sy = math.sin(math.rad(cam.tiltz))
		cam.cy = math.cos(math.rad(cam.tiltz))
	end
	if cam.ptx ~= cam.tiltx then
		cam.sz = math.sin(math.rad(cam.tiltx))
		cam.cz = math.cos(math.rad(cam.tiltx))
	end
	cam.ptx = cam.tiltx
	cam.pty = cam.tilty
	cam.ptz = cam.tiltz
end

function threeDeeWorld:setCameraPosition(x, y, z, tilt)
	self.camera.x = x
	self.camera.y = y
	self.camera.z = z
	self.camera.tilty = tilt or self.camera.tilty
end

--  dP""b8  dP"Yb  88     88     88 .dP"Y8 88  dP"Yb  88b 88
-- dP   `" dP   Yb 88     88     88 `Ybo." 88 dP   Yb 88Yb88
-- Yb      Yb   dP 88  .o 88  .o 88 o.`Y8b 88 Yb   dP 88 Y88
--  YboodP  YbodP  88ood8 88ood8 88 8bodP' 88  YbodP  88  Y8
local collisionType = {
	[-99] = "vtree",
	[-8] = "swamp",
	[-7] = "sand",
	[-6] = "door_u",
	[-5] = "door_d",
	[-4] = "tree",
	[-3] = "bridge_h",
	[-2] = "bridge_v",
	[-1] = "village",
	[0] = "grass",
	[1] = "mountain",
	[2] = "water",
}
function threeDeeWorld:collisionType(t)
	return collisionType[t]
end

function threeDeeWorld:setupCollisionGrid(collisionGrid, collisionList)
	local _timer = tmr.getTime()

	local collisionList = collisionList or {}
	local otherTiles = {}
	self.colMap = {}
	local markedGrid = {}
	for x = 0, #collisionGrid do
		self.colMap[x] = {}
		markedGrid[x] = {}
		for y = 0, #collisionGrid[1] do
			local ny = #collisionGrid[1]-y
			local t = collisionGrid[x][ny]
			self.colMap[x][y] = t
		end
	end

	for x = 0, #collisionGrid do
		for y = 0, #collisionGrid[1] do
			if self.colMap[x][y] > 0 then
				markedGrid[x][y] = false
			elseif self.colMap[x][y] < 0 then
				otherTiles[#otherTiles+1] = { x = x, y = y, col = self.colMap[x][y] }
			end
		end
	end

	if self.loadedMap == "worldmap" then self.forceCreateCache = false end
	if lfs.exists("mapcache/"..self.loadedMap..".lua") and not self.forceCreateCache then
		self.boxList = lfs.load("mapcache/"..self.loadedMap..".lua")()
	else
		local boxSizeList = {
			{128,24}, {32,128}, {64,12}, {12,64}, {32,12}, {12,32}, {12,12}, {8,8}, {7,7}, {6,6}, {6,2}, {5,5}, {5,4}, {5,3},
			{5,2}, {4,4}, {2,4}, {4,5}, {4,2}, {4,3}, {4,2}, {3,3}, {3,4}, {1,32}, {32,1}, {7,1}, {6,1}, {4,1}, {2,3}, {1,4}, {3,1}, {1,3}, {2,2}, {2,1}, {1,2}, {1,1}
		}
		self.boxList = {}

		for a = 1, #boxSizeList do
			local bw, bh = boxSizeList[a][1], boxSizeList[a][2]
			local count = bw * bh
			for x = 0, #self.colMap-bw+1 do
				for y = 0, #self.colMap[1]-bh+1 do
					local _count = 0
					for bx = x, x + (bw-1) do
						for by = y, y + (bh-1) do
							if not markedGrid[bx][by] then
								if self.colMap[bx][by] > 0 then
									_count = _count + 1
								end
							end
						end
					end
					if _count == count then
						local passthrough = false
						-- if math.random(0,99) == 0 then passthrough = true end
						self.boxList[#self.boxList+1] = { x = x, y = y, w = bw, h = bh, passthrough = passthrough }

						for bx = x, x + (bw-1) do
							for by = y, y + (bh-1) do
								markedGrid[bx][by] = true
							end
						end
					end
				end
			end
		end

		for i = 1, #otherTiles do
			local ot = otherTiles[i]
			if ot.col == -2 then
				self.boxList[#self.boxList+1] = { x = ot.x, y = ot.y, w = 0.2, h = 1 }
				self.boxList[#self.boxList+1] = { x = ot.x+0.8, y = ot.y, w = 0.2, h = 1 }
			elseif ot.col == -3 then
				self.boxList[#self.boxList+1] = { x = ot.x, y = ot.y, w = 1, h = 0.2 }
				self.boxList[#self.boxList+1] = { x = ot.x, y = ot.y+0.8, w = 1, h = 0.2 }
			elseif ot.col == -5 then
				self.boxList[#self.boxList+1] = { x = ot.x-0.1, y = ot.y, w = 0.25, h = 1 }
				self.boxList[#self.boxList+1] = { x = ot.x+0.85, y = ot.y, w = 0.25, h = 1 }
				self.boxList[#self.boxList+1] = { x = ot.x, y = ot.y+0.5, w = 1, h = 0.5 }
			elseif ot.col == -6 then
				self.boxList[#self.boxList+1] = { x = ot.x, y = ot.y, w = 0.2, h = 1 }
				self.boxList[#self.boxList+1] = { x = ot.x+0.8, y = ot.y, w = 0.2, h = 1 }
				self.boxList[#self.boxList+1] = { x = ot.x+0.2, y = ot.y, w = 0.6, h = 0.5 }
			elseif ot.col == -99 then
				self.boxList[#self.boxList+1] = { x = ot.x+0.35, y = ot.y+0.1, w = 0.3, h = 0.25 }
			end
		end

		-- self.boxList[#self.boxList+1] = { x = -1, y = -1, w = 1, h = #collisionGrid[1]+2, worldEdge = true }
		-- self.boxList[#self.boxList+1] = { x = #collisionGrid+1, y = -1, w = 1, h = #collisionGrid[1]+2, worldEdge = true }
		-- self.boxList[#self.boxList+1] = { x = -1, y = -1, w = #collisionGrid+2, h = 1, worldEdge = true }
		-- self.boxList[#self.boxList+1] = { x = -1, y = #collisionGrid+1, w = #collisionGrid+2, h = 1, worldEdge = true }

		for i = 1, #collisionList do
			self.boxList[#self.boxList+1] = collisionList[i]
		end

		local txt = "return {\n"
		for i = 1, #self.boxList do
			txt = txt .. "\t[" .. i .. "] = { x = " .. self.boxList[i].x .. ", y = " .. self.boxList[i].y .. ", w = " .. self.boxList[i].w .. ", h = " .. self.boxList[i].h .. "},\n"
		end
		txt = txt .. "}"
		lfs.write("mapcache/"..self.loadedMap..".lua", txt, all)
	end

	for i = 1, #self.boxList do
		self.colWorld:add(self.boxList[i], self.boxList[i].x * self.tileSize, self.boxList[i].y * self.tileSize, self.boxList[i].w * self.tileSize, self.boxList[i].h * self.tileSize)
	end

	self.gridCanvas = lgr.newCanvas(math.min((#collisionGrid+1)*self.tileSize/4,8192), math.min((#collisionGrid[1]+1)*self.tileSize/4,8192))

	self.gridCanvas:renderTo(function()
		local s = math.min(0.25, self.gridCanvas:getWidth() / ((#collisionGrid+1)*self.tileSize/4))
		local buff = 4
		if self.loadedMap == "worldmap" then buff = 32 end
		lgr.push()
		lgr.scale(s)
		for i, c in pairs(self.boxList) do
			local l = self.colWorld:getRect(c)
			lgr.setColor(0,100,255)
			lgr.rectangle("fill", l.l, l.t, l.w, l.h)
			lgr.setColor(0,0,100)
			lgr.rectangle("fill", l.l + buff, l.t, l.w - buff, l.h - buff)
			if c.passthrough then
				lgr.setColor(200,0,0)
			else
				lgr.setColor(0,0,200)
			end
			lgr.rectangle("fill", l.l + buff, l.t + buff, l.w - (buff * 2), l.h - (buff * 2))
		end
		lgr.pop()
	end)

	print("Collision Box Count:", #self.boxList)
	print("It took " .. tmr.getTime() - _timer .. " seconds to generate.")
end


-- 888888 88  88 88""Yb 888888 888888     8888b.  888888 888888     888888 88   88 88b 88  dP""b8 888888 88  dP"Yb  88b 88 .dP"Y8
--   88   88  88 88__dP 88__   88__        8I  Yb 88__   88__       88__   88   88 88Yb88 dP   `"   88   88 dP   Yb 88Yb88 `Ybo."
--   88   888888 88"Yb  88""   88""        8I  dY 88""   88""       88""   Y8   8P 88 Y88 Yb        88   88 Yb   dP 88 Y88 o.`Y8b
--   88   88  88 88  Yb 888888 888888     8888Y"  888888 888888     88     `YbodP' 88  Y8  YboodP   88   88  YbodP  88  Y8 8bodP'
function threeDeeWorld:point3dto2d(x, y, z, cam)
	local tt = tmr.getTime()
	local t = self.timer
	local dw = tonumber(screen.drawWidth or lgr.getWidth())
	local dh = tonumber(screen.drawHeight or lgr.getHeight())
	local wss = self.zoomScale
	local xy, xz, yz, yx, zx, zy, scale, dist, rx, ry, rz, xo, yo

	local xoff = 0
	local yoff = 0
	local zoff = 0
	-- if y == 32 and z - cam.locZ < 1700 then
	-- 	yoff = math.cos(x + t * 0.5) * math.sin(z + t * 1.7) * 4
	-- end
	if y == 31 then
		yoff = math.cos(x + t * 0.5) * math.sin(z + t * 1.7) * 2
	end
	if y == 33 then
		-- xoff = math.cos(x + t * 0.3) * math.sin(z + t * 0.5) * 10
		zoff = math.cos(x + t * 0.4) * math.sin(z + t * 1) * 10
		yoff = math.cos(x + t * 0.5) * math.sin(z + t * 1.7) * 2
	end
	if y == 60 then
		xoff = math.cos(x + t * 0.5) * math.sin(z + t * 1.7) * 2
		zoff = math.cos(x + t * 1) * math.sin(z + t * 0.7) * 5
	end
	if y == 40 then
		xoff = math.cos(x + t * 0.5) * math.sin(z + t * 1.7) * 2
		zoff = math.cos(x + t * 1) * math.sin(z + t * 0.7) * 5
		yoff = math.cos(x + t * 1.03) * math.sin(z + t * 0.6) * 10
	end

	--3D Calculations are done here!!!!
	rx = -cam.locX + x - xoff
	ry = -cam.y + y - yoff
	rz = -cam.locZ + z - zoff

	xy = cam.cx * ry - cam.sx * rz
	xz = cam.sx * ry + cam.cx * rz

	yz = cam.cy * xz - cam.sy * rx
	yx = cam.sy * xz + cam.cy * rx

	zx = cam.cz * yx - cam.sz * xy
	zy = cam.sz * yx + cam.cz * xy

	scale = self.focal / (self.focal + yz)
	dist = yz
	xo = dw / 2
	yo = dh / 2
	rx = xo + (zx * scale) * wss
	ry = yo + (zy * scale) * wss + ((dist > 1000 and cam.tilty < 30 and self.bendHorizon) and ((dist - 1000) / self.curveAmount * self.curvature) or 0)

	performanceGraph:addToCalcTime(tmr.getTime() - tt)
	if dist < -self.focal then
		return huge, huge, 0, 0, false
	end

	return rx, ry, dist, scale * wss, (rx >= 0 and rx <= dw and ry >= 0 and ry <= dh)
end

function threeDeeWorld:backface(x1, y1, x2, y2, x3, y3, x4, y4)
	if not x1 or not x2 or not x3 or not x4 or not y1 or not y2 or not y3 or not y4 then return false end
	return not ((y3-y1)/(x3-x1) - (y2-y1)/(x2-x1) <= 0) ~= ((x1 <= x3) == (x1 > x2))
end

function threeDeeWorld:onScreen(p1, p2, p3, p4, w, h, av)
	p4 = p4 or p1
	local buff = 0
	local minDist = -self.focal
	local maxDist = self.maxDistance + self.chunkSize
	if not p1.cx or not p2.cx or not p3.cx or not p4.cx or not p1.cy or not p2.cy or not p3.cy or not p4.cy then
		return false
	elseif (p1.dist > minDist or p2.dist > minDist or p3.dist > minDist or p4.dist > minDist) and av then
		return true
	elseif  (p1.dist < minDist or p2.dist < minDist or p3.dist < minDist or p4.dist < minDist) or
			(p1.dist > maxDist or p2.dist > maxDist or p3.dist > maxDist or p4.dist > maxDist) then
		return false
	elseif  (p1.cy < 0 and p2.cy < 0 and p3.cy < 0 and p4.cy < 0) or
			(p1.cy > w and p2.cy > w and p3.cy > w and p4.cy > w) or
			(p1.cx < -buff and p2.cx < -buff and p3.cx < -buff and p4.cx < -buff) or
			(p1.cx > h + buff and p2.cx > h + buff and p3.cx > h + buff and p4.cx > h + buff) then
		return false
	else
		return true
	end
end


-- 88      dP"Yb     db    8888b.      8b    d8    db    88""Yb
-- 88     dP   Yb   dPYb    8I  Yb     88b  d88   dPYb   88__dP
-- 88  .o Yb   dP  dP__Yb   8I  dY     88YbdP88  dP__Yb  88"""
-- 88ood8  YbodP  dP""""Yb 8888Y"      88 YY 88 dP""""Yb 88
function threeDeeWorld:loadLevel(map)
	print("Loading " .. map)

	self.colWorld = nil
	collectgarbage()

	local pl, sl, cs, ws, cg, cl, ml = lfs.load("maps/"..map..".lua")()
	local timeTaken = tmr.getTime()

	self.loadedMap = map

	self.showMapWelcome = ws.showMapWelcome or false

	self.tileSize = 128

	self.colWorld = bump.newWorld(512)

	game:playMusic(ws.music)

	self.loaded = false

	self.chunkSize = 512

	self.entryPoints = {}
	local entry = ws.entryPoints or {}
	for i, e in pairs(entry) do
		print("Entry:", i, e.x, e.z, e.facing)
		self.entryPoints[i] = { x = e.x or 0, y = e.y or 0, z = e.z or 0, facing = e.facing or 0 }
	end

	self.warpPoints = {}
	local warp = ws.warpPoints or {}
	for i, e in pairs(warp) do
		local x, y, z, w, h = e.x * TILESIZE or 0, e.y, e.z * TILESIZE or 0, e.w * TILESIZE or 0, e.h * TILESIZE or 0
		print("Warp:", i, x, z, w, h, e.map, e.entryPoint)
		self.warpPoints[i] = { x = x, y = y, z = z, w = w, h = h, map = e.map or nil, entryPoint = e.entryPoint or nil, action = e.action or nil, entity = e.entity or nil }
	end

	self.ambientDim = ws.dim or 1
	self.levelHasMap = ws.hasMap or false

	self.cameraDistance = cs.cameraDistance or nil
	self.cameraHeight = cs.cameraHeight or 250
	self.cameraLocked = cs.locked or false
	if self.cameraLocked then
		self.cameraLockedTo = { x = cs.x, y = cs.y, z = cs.z }
		-- print(cs.x, cs.y, cs.z)
	end


	self.lightingCanvas = lgr.newCanvas(screen.drawWidth, screen.drawHeight)

	self.focal = 400

	self.points = {}
	self.polys = {}
	self.sprites = {}

	self.skyColor = {}
	self.cloudColor = {}
	self.fogColor = {}
	self.fogIntensity = 0

	self.calcCamRot = true

	self.viewTilt = 0

	self.threeDeeSprites = false
	self.drawTextures = true

	self.pointCache = {}

	self.disableEnvironment = ws.disableEnvironment or false
	self.backgroundColor = ws.backgroundColor or {0,0,0}

	self.mapName = ws.mapName

	self.bendHorizon = ws.bendHorizon or false
	self.loopEdges = ws.loopEdges or false
	self.worldWidth = ws.width
	self.worldDepth = ws.depth
	self.tilesWide = ws.tilesWide
	self.tilesDeep = ws.tilesDeep

	if ml then
		print("Models List")
		for i, m in pairs(ml) do
			local model = lfs.load("models/"..m.file..".lua")()
			local x, y, z = m.x or 0, m.y or 0, m.z or 0
			local scale = m.scale or 1
			print(m.file, x, y, z, scale)
			for j, p in pairs(model.polys) do
				local n = #pl+1
				pl[n] = { points = {}, texture = p.tex, color = p.color, layer = p.layer, sortPoint = p.sortPoint, flip = p.flip }
				for k = 1, #p.pts do
					local px, py, pz = (model.points[p.pts[k]][1] * scale) + x, (model.points[p.pts[k]][2] * scale) + y, (model.points[p.pts[k]][3] * scale) + z
					pl[n].points[k] = { px, py, pz }
				end
			end

			print("Are there collisions?")
			if model.collisions then
				print("Collisions")
				for j = 1, #model.collisions do
					local mc = model.collisions[j]
					local px, pz = (mc.x * scale) + x / 128, (mc.z * scale) + z / 128
					local sw, sd = (mc.w * scale) / 128, (mc.d * scale) / 128
					print(px, pz, sw, sd)
					cl[#cl+1] = { x = px, y = pz, w = sw, h = sd }
				end
			end
		end
	end

	for i = 1, #pl do
		self:addPolygon(pl[i].points, pl[i].color, pl[i].texture, pl[i].layer, pl[i].flip, pl[i].sortPoint, pl[i].dim or 0)
	end

	for i, n in pairs(game.npcMasterList) do
		local e = game.entity[i]
		if e.map == self.loadedMap then
			print("Adding Sprite to World for " .. i)
			sl[#sl+1] = { image = e.spriteID, x = 400, y = 0, z = 600, name = e.spriteID, scale = CHARACTERSCALE, absoluteScale = false, layer = 0, angle = 90, worldSprite = true, shadow = true }
		end
	end

	-- Load Entities and Sprites
	self.entity = {}
	for i = 1, #sl do
		if sl[i].entity then
			-- print(sl[i].name)
			sl[i].parentWorld = self
			self.entity[sl[i].name] = scenery[sl[i].entity]:create(sl[i])
			sl[i].image = self.entity[sl[i].name].image
			self:addSprite(sl[i])
		else
			self:addSprite(sl[i])
			if sl[i].shadow then
				self:addSprite { name = "shadow_"..sl[i].name, x = sl[i].x, y = sl[i].y, z = sl[i].z, image = "shadow", scale = sl[i].scale * 0.6 * (sl[i].shadowScale or 1),
									absoluteScale = false, layer = -0.00001, worldSprite = true, forceThreeDee = true }
			end
		end
	end

	-- self:addSprite { image = "crate", x = 400, y = 0, z = 600, name = self.nameID.."_sprite" }

	self.maxDistance = 1700
	self.fieldOfView = 1200
	self.fogDistance = 1500

	self.chunkList = {}
	self.chunksWide = 0
	self.chunksLong = 0

	print("Poly Count:", #self.polys)
	for i = 1, #self.polys do
		local chunksThisPolyIsIn = {}
		-- print("Polygon #"..i)
		for j = 1, #self.polys[i].points do
			local p = self.polys[i].points[j]
			local x, z = math.floor(self.points[p].x / self.chunkSize), math.floor(self.points[p].z / self.chunkSize)
			if x > self.chunksWide then self.chunksWide = x end
			if z > self.chunksLong then self.chunksLong = z end
			chunksThisPolyIsIn[j] = { x = x, z = z }
		end

		local compressedChunkList = {}
		for j = 1, #chunksThisPolyIsIn do
			local c = "c" .. chunksThisPolyIsIn[j].x .. "_" .. chunksThisPolyIsIn[j].z
			compressedChunkList[c] = true
			if not self.chunkList[c] then
				self.chunkList[c] = {
					points = {},
					polys = {},
					sprites = {},
					dbgColor = {math.random(100,200),math.random(100,200),math.random(100,200)}
				}
			end
		end

		local chunks = {}
		for j, c in pairs(compressedChunkList) do
			chunks[#chunks+1] = j
		end
		table.sort(chunks)
		self.polys[i].chunk = chunks[1]
	end

	-- table.sort(self.chunkList)

	for i, p in pairs(self.polys) do
		local cpn = #self.chunkList[p.chunk].polys+1
		self.chunkList[p.chunk].polys[cpn] = i
	end

	print(self.chunksWide, self.chunksLong)
	for x = 0, self.chunksWide do
		for z = 0, self.chunksLong do
			local chunk = "c"..x.."_"..z
			local listOfPointsInThisChunk = {}
			for i, p in pairs(self.chunkList[chunk].polys) do
				-- print(i, p)
				for j = 1, #self.polys[p].points do
					-- print(self.polys[p].points[j])
					listOfPointsInThisChunk[self.polys[p].points[j]] = true
				end
			end

			for i, pic in pairs(listOfPointsInThisChunk) do
				self.chunkList[chunk].points[#self.chunkList[chunk].points+1] = i
			end
		end
	end

	self.globalSprites = {}
	self.globalPoints = {}
	for i, s in pairs(self.sprites) do
		if not s.worldSprite then
			local x, z = math.floor(self.points[s.point].x / self.chunkSize), math.floor(self.points[s.point].z / self.chunkSize)
			local chunk = "c"..x.."_"..z
			local j = #self.chunkList[chunk].sprites+1
			self.chunkList[chunk].sprites[j] = i
			self.chunkList[chunk].points[#self.chunkList[chunk].points+1] = s.point
		else
			self.globalSprites[#self.globalSprites+1] = i
			self.globalPoints[#self.globalPoints+1] = s.point
		end
	end

	if game.worldMapPos then
		self.colWorld:add(game.worldMapPos, game.worldMapPos.x, game.worldMapPos.z, game.worldMapPos.w, game.worldMapPos.h)
	end

	self:setupCollisionGrid(cg, cl)

	self:generateMiniMap()

	self:setupCamera(cs)

	initializeAnimations()

	self.welcomeScreenFade = 0
	self.welcomeScreenWidth = 0
	self.welcomeScreenTextOffset = -20
	self.welcomeScreenBackFade = 1

	if self.showMapWelcome then
		cron.after(0.5, function()
			self.fadingTween = flux.to(self, 0.5, { welcomeScreenFade = 1 }):ease("linear"):delay(0.5)
			self.openingTween = flux.to(self, 0.85, { welcomeScreenWidth = 1 }):ease("quadout")
			self.textmoverTween = flux.to(self, 0.5, { welcomeScreenTextOffset = 0 }):ease("quadout"):delay(0.5):oncomplete(function()
				-- self.textmoverTween = flux.to(self, 2, { welcomeScreenTextOffset = 5 }):ease("linear")
				cron.after(2, function()
					self.fadingTween = flux.to(self, 0.5, { welcomeScreenFade = 0 }):ease("linear")
					self.fadingTweenB = flux.to(self, 0.5, { welcomeScreenBackFade = 0 }):ease("linear")
					self.textmoverTween = flux.to(self, 0.5, { welcomeScreenTextOffset = 20 }):ease("quadout"):oncomplete(function()
						self.showMapWelcome = false
					end)
				end)
			end)
		end)
	end

	print("Generated Level Geometry in " .. (tmr.getTime() - timeTaken) .. " seconds.")
	collectgarbage()

	self.loaded = true
end

function threeDeeWorld:drawWelcomeSign()
	if not self.showMapWelcome then return false end
	screen.overlay:renderTo(function()
		local t = self.mapName
		lgr.push()
		lgr.translate(0, 40)
		lgr.scale(self.welcomeScreenWidth, 1)
		lgr.setColor(0, 0, 0, 160 * self.welcomeScreenBackFade)
		lgr.rectangle("fill", 0, 0, screen.overlay:getWidth(), 30)
		lgr.pop()

		lgr.push()
		lgr.translate(0, 40)
		lgr.scale(1)
		lgr.setFont(font.large)
		lgr.setColor(0, 200, 200, self.welcomeScreenFade * 200)
		lgr.printf(t, self.welcomeScreenTextOffset + 1, 10, screen.overlay:getWidth() / 1, "center")
		lgr.setColor(255, 255, 255, self.welcomeScreenFade * 255)
		lgr.printf(t, self.welcomeScreenTextOffset, 8, screen.overlay:getWidth() / 1, "center")
		lgr.pop()
	end)
end

-- Yb        dP    db    88""Yb 88""Yb     88""Yb  dP"Yb  88 88b 88 888888     88  88    db    88b 88 8888b.  88     888888 88""Yb
--  Yb  db  dP    dPYb   88__dP 88__dP     88__dP dP   Yb 88 88Yb88   88       88  88   dPYb   88Yb88  8I  Yb 88     88__   88__dP
--   YbdPYbdP    dP__Yb  88"Yb  88"""      88"""  Yb   dP 88 88 Y88   88       888888  dP__Yb  88 Y88  8I  dY 88  .o 88""   88"Yb
--    YP  YP    dP""""Yb 88  Yb 88         88      YbodP  88 88  Y8   88       88  88 dP""""Yb 88  Y8 8888Y"  88ood8 888888 88  Yb
function threeDeeWorld:getEntryPos(name)
	if not self.entryPoints[name] then return false end
	return self.entryPoints[name].x, self.entryPoints[name].z, self.entryPoints[name].facing
end

function threeDeeWorld:checkForWarpPoint(x, z)
	for i, w in pairs(self.warpPoints) do
		if overlap(x, z, 1, 1,  w.x, w.z, w.w, w.h) then
			if w.map then
				return i, "changeMap", { map = w.map, entryPoint =  w.entryPoint }
			elseif w.action then
				return i, w.action, w.entity
			end
		end
	end
	return "", "", nil
end

-- 8b    d8 88 88b 88 88     8b    d8    db    88""Yb
-- 88b  d88 88 88Yb88 88     88b  d88   dPYb   88__dP
-- 88YbdP88 88 88 Y88 88     88YbdP88  dP__Yb  88"""
-- 88 YY 88 88 88  Y8 88     88 YY 88 dP""""Yb 88
function threeDeeWorld:generateMiniMap()
	self.mapCanvasDraw = lgr.newCanvas(64+8, 64+8)
	self.mapCanvasDraw:clear(0,0,0,0)

	self.mapCanvas = lgr.newCanvas(self.tilesWide, self.tilesDeep)
	self.mapCanvas:clear(0,0,0,100)
	lgr.setCanvas(self.mapCanvas)
	for x = 0, self.tilesWide - 1 do
		for z = 0, self.tilesDeep - 1 do
			if self.colMap[x][z] == 1 then
				lgr.setColor(136,125,87)
			elseif self.colMap[x][z] == 2 then
				lgr.setColor(100,130,255)
			elseif self.colMap[x][z] == -1 then
				lgr.setColor(255,0,255)
			elseif self.colMap[x][z] == -2 or self.colMap[x][z] == -3 then
				lgr.setColor(255,255,0)
			elseif self.colMap[x][z] == -4 then
				lgr.setColor(0,150,0)
			elseif self.colMap[x][z] == -7 then
				lgr.setColor(255,255,100)
			elseif self.colMap[x][z] == -8 then
				lgr.setColor(140,0,140)
			else
				lgr.setColor(0,200,0)
			end
			lgr.rectangle("fill", x, self.tilesDeep-z-1, 1, 1)
		end
	end
	-- for x = 0, self.tilesWide - 1 do
	-- 	for z = 0, self.tilesDeep - 1 do
	-- 		if self.colMap[x][z] ~= 0 then
	-- 			if self.colMap[x][z] == -1 then
	-- 				lgr.setColor(150,0,150)
	-- 				lgr.rectangle("fill", x-1, z-1, 3, 3)
	-- 				lgr.setColor(255,0,255)
	-- 				lgr.rectangle("fill", x, z, 1, 1)
	-- 			end
	-- 		end
	-- 	end
	-- end
	lgr.setCanvas()
end

function threeDeeWorld:drawMiniMap()
	if game.displayMap and self.levelHasMap then
		local s = 1
		local mapSize = 64
		local px, py = math.floor(game.worldMapPos.x / 128 * s), math.floor((self.worldDepth - game.worldMapPos.z) / 128 * s)
		local ax, ay = math.clamp(math.floor(game.worldMapPos.airshipX / 128 * s), px - (mapSize / 2), px + (mapSize / 2)), math.clamp(math.floor((self.worldDepth - game.worldMapPos.airshipZ) / 128 * s), py - (mapSize / 2), py + (mapSize / 2))
		local ox, oy = px - (mapSize / 2), py - (mapSize / 2)

		lgr.setCanvas(self.mapCanvasDraw)
		self.mapCanvasDraw:clear(0,0,0,0)

		lgr.push()
		lgr.translate(4, 4)

		lgr.translate(-ox, -oy)

		lgr.setScissor(4, 4, 64, 64)

		lgr.setColor(255,255,255)
		lgr.draw(self.mapCanvas, 0, 0, 0, 1, 1)

		lgr.setScissor()

		if self.loadedMap == "worldmap" then
			lgr.setColor(0,255,0)
			lgr.circle("fill", ax, ay, 2)
			lgr.setColor(0,0,255)
			lgr.circle("fill", ax, ay, 1)
		end

		lgr.setColor(255,255,0)
		lgr.circle("fill", px, py, 2)
		lgr.setColor(255,0,0)
		lgr.circle("fill", px, py, 1)

		lgr.pop()

		lgr.setCanvas()

		screen.overlay:renderTo(function()
			lgr.setColor(255,255,255)
			lgr.draw(self.mapCanvasDraw, screen.overlay:getWidth() - (64 + 24), screen.overlay:getHeight() - (64 + 24), 0, 1, 1, 4, 4)
		end)
	end
end

-- 8888b.  88""Yb    db    Yb        dP     8888b.  888888 88""Yb 88   88  dP""b8     88 88b 88 888888  dP"Yb
--  8I  Yb 88__dP   dPYb    Yb  db  dP       8I  Yb 88__   88__dP 88   88 dP   `"     88 88Yb88 88__   dP   Yb
--  8I  dY 88"Yb   dP__Yb    YbdPYbdP        8I  dY 88""   88""Yb Y8   8P Yb  "88     88 88 Y88 88""   Yb   dP
-- 8888Y"  88  Yb dP""""Yb    YP  YP        8888Y"  888888 88oodP `YbodP'  YboodP     88 88  Y8 88      YbodP
function threeDeeWorld:drawDebugInfo()
	if not performanceGraph.enabled then return false end
	screen.overlay:renderTo(function()
		local cam = self.camera
		lgr.push()
		lgr.origin()
		lgr.setColor(255,255,255)

		local t = ""
		t = t .. "Loaded Map:\n"
		t = t .. "Dimensions:\n"
		t = t .. "Camera:\n"
		t = t .. "\n"
		if game.loaded then
			t = t .. "Player Pos:\n"
			t = t .. "Player Tile:\n"
			if game_worldmap.world.loadedMap == "worldmap" then
				t = t .. "Airship Pos:\n"
				t = t .. "Flight Elev:\n"
			end
			t = t .. "Collision:\n"
			t = t .. "Cur Trigger:\n"
		end
		t = t .. "World Pts:\n"
		t = t .. "Points:\n"
		t = t .. "Textures:\n"
		t = t .. "Sprites:\n"
		t = t .. "Col Box Count:\n"
		t = t .. "View Width:\n"
		t = t .. "Max Dist:\n"
		if game.loaded then
			t = t .. "Game Timer:\n"
		end
		t = t .. "Time of Day:\n"
		t = t .. "Sky Color:\n"
		t = t .. "Fog Color:\n"
		lgr.setFont(font.tiny)
		lgr.prints(t, 8, 8)

		local t = ""
		t = t .. self.mapName .. " (" .. self.loadedMap .. ".lua)\n"
		t = t .. self.tilesWide .. " x " .. self.tilesDeep .. "\n"
		t = t .. "X: " .. math.round(cam.x,2) .. ", Y: " .. math.round(cam.y,2) .. ", Z: " .. math.round(cam.z,2) .. "\n"
		t = t .. "Dist: " .. tostring(self.cameraDistance) .. ", Height: " .. tostring(self.cameraHeight) .. ", Tilt: " .. tostring(cam.tilty) .. "\n"
		if game.loaded then
			local colType = self.colMap[math.floor(game.worldMapPos.x / 128)][math.floor(game.worldMapPos.z / 128)]
			local colName = collisionType[colType]
			t = t .. math.floor(game.worldMapPos.x) .. ", " .. math.floor(game.worldMapPos.z) .. "\n"
			t = t .. math.floor(game.worldMapPos.x / 128) .. ", " .. (game_worldmap.world.tilesDeep - math.floor(game.worldMapPos.z / 128)) .. " (" .. (math.floor(game.worldMapPos.z / 128)) .. ")\n"
			if game_worldmap.world.loadedMap == "worldmap" then
				t = t .. math.floor(game.worldMapPos.airshipX / 128) .. ", " .. math.floor(game.worldMapPos.airshipZ / 128) .. "\n"
				t = t .. math.floor(game_worldmap.flightElevation) .. "\n"
			end
			t = t .. "Type: " .. colType .. " (" .. colName .. ")\n"
			t = t .. tostring(game_worldmap.triggerName) .. "\n"
		end
		t = t .. #self.points .. "\n"
		t = t .. calculatedPoints .. "\n"
		t = t .. drawnTextures .. "\n"
		t = t .. drawnSprites .. "\n"
		t = t .. #self.boxList .. "\n"
		t = t .. self.fieldOfView .. "\n"
		t = t .. self.maxDistance .. "\n"
		if game.loaded then
			t = t .. formatTime(game.timer) .. " (" .. math.round(game.timer,2) .. ")" .. "\n"
		end
		t = t .. math.floor(self.timeOfDay) .. " (" .. self.timeOfDaySeg .. ")  Day #" .. self.dayCount .. "\n"
		t = t .. math.floor(self.skyColor[1]) .. ", " .. math.floor(self.skyColor[2]) .. ", " .. math.floor(self.skyColor[3]) .. "\n"
		t = t .. math.floor(self.fogColor[1]) .. ", " .. math.floor(self.fogColor[2]) .. ", " .. math.floor(self.fogColor[3]) .. "\n"
		lgr.setFont(font.tiny)
		lgr.prints(t, 80, 8)

		-- t = "Chunks:\n"
		-- for i = 1, #self.chunksDrawnThisFrame do
		-- 	t = t .. " " .. self.chunksDrawnThisFrame[i] .. "\n"
		-- end
		-- lgr.prints(t, 8, 130)

		lgr.pop()
	end)
end

function threeDeeWorld:debugDrawHotzones()
	for j, w in pairs(self.warpPoints) do
		local n = math.random(#self.points+1,999999999)
		local newPoints = { {w.x, w.y, w.z + w.h, n}, {w.x + w.w, w.y, w.z + w.h, n+0.1}, {w.x + w.w, w.y, w.z, n+0.2}, {w.x, w.y, w.z, n+0.3} }

		for j = 1, #newPoints do
			local cx, cy, dist, scale, onScreen = self:point3dto2d(newPoints[j][1], newPoints[j][2], newPoints[j][3], self.camera)
			self.worldPoints[newPoints[j][4]] = { cx = cx, cy = cy, dist = dist, scale = scale, onScreen = onScreen }
			calculatedPoints = calculatedPoints + 1
		end

		local t
		if w.map then
			t = "Warp to Map: " ..w.map .. "\n" .. "Entry: " .. w.entryPoint
		else
			t = j .. "\nAction: " .. w.action
		end
		self.worldObjects[#self.worldObjects+1] = {
			points = {
				self.worldPoints[newPoints[1][4]].cx, self.worldPoints[newPoints[1][4]].cy,
				self.worldPoints[newPoints[2][4]].cx, self.worldPoints[newPoints[2][4]].cy,
				self.worldPoints[newPoints[3][4]].cx, self.worldPoints[newPoints[3][4]].cy,
				self.worldPoints[newPoints[4][4]].cx, self.worldPoints[newPoints[4][4]].cy,
			}, dist = 1, kind = "poly", color = {255,0,0,100}, border = true, layer = 20, dim = 1, text = t
		}
		if w.action == "triggerEntity" then self.worldObjects[#self.worldObjects].color = {0,255,0,100} end
	end

	for j, w in pairs(self.entryPoints) do
		local n = math.random(#self.points+1,999999999)
		local x, y, z = w.x * TILESIZE + TILESIZE / 2, w.y, w.z * TILESIZE + TILESIZE / 2
		local newPoints = { {x-8, y, z+8, n}, {x+8, y, z+8, n+0.1}, {x+8, y, z-8, n+0.2}, {x-8, y, z-8, n+0.3} }

		for j = 1, #newPoints do
			local cx, cy, dist, scale, onScreen = self:point3dto2d(newPoints[j][1], newPoints[j][2], newPoints[j][3], self.camera)
			self.worldPoints[newPoints[j][4]] = { cx = cx, cy = cy, dist = dist, scale = scale, onScreen = onScreen }
			calculatedPoints = calculatedPoints + 1
		end

		self.worldObjects[#self.worldObjects+1] = {
			points = {
				self.worldPoints[newPoints[1][4]].cx, self.worldPoints[newPoints[1][4]].cy,
				self.worldPoints[newPoints[2][4]].cx, self.worldPoints[newPoints[2][4]].cy,
				self.worldPoints[newPoints[3][4]].cx, self.worldPoints[newPoints[3][4]].cy,
				self.worldPoints[newPoints[4][4]].cx, self.worldPoints[newPoints[4][4]].cy,
			}, dist = 0, kind = "poly", color = {0,0,255,100}, border = true, layer = 20, dim = 1, text = "Name: " .. j
		}
	end
end
