local lgr = love.graphics

local _transition = {}

function _transition:init(arg)
	print("Initializing Transition")
	local arg = arg or {}
	self.type = arg.type or "normal"
	self.opacity = 0
	self.colorComponents = { r = arg.r or 0, g = arg.g or 0, b = arg.b or 0 }
	self.speed = arg.speed or 0.25

	self._x = 0
	self._y = 0
	self._w = 0
	self._h = 0

	self.ease = arg.ease or "linear"
	self.tweens = {
		o = nil
	}
end

function _transition:update(dt)
	self._w = lgr.getWidth()
	self._h = lgr.getHeight()
	self.color = { self.colorComponents.r, self.colorComponents.g, self.colorComponents.b, math.min(self.opacity,1) * 255 }
end

function _transition:draw()
	if self.opacity > 0 then
		if self.type == "fade" then
			lgr.setColor(self.color)
			lgr.rectangle("fill", self._x, self._y, self._w, self._h)
		end
	end
	-- lgr.setColor(0,255,0)
	-- lgr.print(self.opacity, 0, 0)
end

function _transition:isFinished()
	return self.opacity == 0
end

function _transition:setSpeed(s)
	self.speed = s or 0.25
end

function _transition:fadeOut(func, s)
	local func = func or function() end
	self.type = "fade"
	-- self.tweens.o:stop()
	self.tweens.o = flux.to(self, self.speed, { opacity = 1.1 }):ease(self.ease):oncomplete(func)
end

function _transition:fadeIn(func, s)
	local func = func or function() end
	self.type = "fade"
	-- self.tweens.o:stop()
	self.tweens.o = flux.to(self, self.speed, { opacity = 0 }):ease(self.ease):oncomplete(func)
end

function _transition:doFadeCycle(func)
	local func = func or function() end
	self:fadeOut(function() self:fadeIn() func() end)
end

function _transition:setOpacity(o)
	self.opacity = o
end

function _transition:setColor(r, g, b)
	self.colorComponents.r = r or 0
	self.colorComponents.g = g or 0
	self.colorComponents.b = b or 0
end

function _transition:toColor(r, g, b, s, e)
	local e = e or "linear"
	local s = s or self.speed
	self.tweens.r:stop()
	self.tweens.g:stop()
	self.tweens.b:stop()
	if r then
		self.tweens.r = flux.to(self.colorComponents, s, { r = r }):ease(e)
	end
	if g then
		self.tweens.g = flux.to(self.colorComponents, s, { g = g }):ease(e)
	end
	if b then
		self.tweens.b = flux.to(self.colorComponents, s, { b = b }):ease(e)
	end
end

return _transition
