local lgr = love.graphics
local win = love.window

messageBoxModule = {}

function messageBoxModule:create(v)
	local _i = {}
	setmetatable(_i, {__index = self})
	_i:setup(v)
	return _i
end
function messageBoxModule:setup(v)
	local v = v or {}
	self.message = v.message or v.msg or v.text or "<<NO TEXT TO DISPLAY>>"
	self.typeSpeed = v.typeSpeed or v.typespeed or v.speed or 100
	self.x = 20
	self.y = 20
	self.z = v.z or 1000000
	self.width = 340
	self.caret = 0
	self.font = v.font or font.dialog

	self.letterSpacing = v.letterSpacing or 1
	self.lineHeight = self.font:getLineHeight()
	self.fontHeight = self.font:getHeight()
	self.lineSpacing = v.lineSpacing or 0

	self.letter = {}
	local lastSpace = 0
	local caretLocation = 0
	local lineTop = 0
	local spaces = {}
	local j = 1
	for i = 1, #self.message do
		local isSpace = false
		local c = self.message:sub(i,i)
		local w = self.font:getWidth(c) - 1
		if c == " " then
			lastSpace = caretLocation
			spaces[#spaces+1] = {
				char = i,
				location = lastSpace + w + self.letterSpacing
			}
		elseif c == "\n" then
			lineTop = lineTop + self.lineHeight * self.fontHeight + self.lineSpacing
			caretLocation = 0
		elseif c == "\t" then
			w = 23
		end
		if c ~= " " and c ~= "\n" then
			-- print(c, w, caretLocation, lastSpace)
			local x = caretLocation
			local y = lineTop
			self.letter[j] = {
				char = c,
				width = w,
				x = x,
				y = y
			}
		end
		caretLocation = caretLocation + w + self.letterSpacing
		j = j + 1
	end



	self.delete = false
end

function messageBoxModule:update(dt)
	self.caret = self.caret + dt * self.typeSpeed
	self.displayMessage = self.message:sub(1, math.floor(self.caret))
end

function messageBoxModule:draw()
	lgr.push()
	lgr.translate(self.x, self.y)

	lgr.setColor(0,0,0,100)
	lgr.rectangle("fill", 0, 0, self.width, 400)

	lgr.setFont(self.font)
	lgr.setColor(255,255,255)
	lgr.printf(self.displayMessage, 0, 0, self.width, self.align)

	for i, c in pairs(self.letter) do
		lgr.print(c.char, c.x, c.y + 100)
	end

	lgr.pop()
end

function messageBoxModule:speedUp(s)
	self.typeSpeed = s or 100
end

function messageBoxModule:close()
	self.delete = true
end
