local lgr = love.graphics
local win = love.window
local lfs = love.filesystem
local kbd = love.keyboard
local tmr = love.timer

game = {}

keyJump = "c"

SHOWNUMBERS = false

function game:load(arg)
	local timer = tmr.getTime()
	print("Loading Game State")
	local arg = arg or {}
	self.loaded = true
	self.backgroundUpdate = false
	self.backgroundDraw = true
	self.message = ""
	self.transition = 0
	self.name = "Test Bed"
	self.width = 640
	self.height = 360
	self.canvas = lgr.newCanvas(self.width, self.height)
	self.canvas:setFilter("linear", "nearest")

	self.map = lfs.load("maps/map1.lua")()
	self.map.width = #self.map.grid
	self.map.height = #self.map.grid[1]
	print(self.map.width, self.map.height)

	self.map.collisionMap = bump.newWorld(128)

	local tempColMap = {}
	local markedGrid = {}
	self.map.plantGrid = {}
	for x = 1, self.map.width do
		tempColMap[x-1] = {}
		markedGrid[x-1] = {}
		self.map.plantGrid[x] = {}
		for y = 1, self.map.height do
			tempColMap[x-1][y-1] = self.map.grid[x][y]
			markedGrid[x-1][y-1] = false
			self.map.plantGrid[x][y] = { item = "" }
		end
	end

	for i = 1, #self.map.plantList do
		local p = self.map.plantList[i]
		self.map.plantGrid[p.x][p.y] = { item = p.item }
	end

	local boxSizeList = {
		{128,24}, {32,128}, {64,12}, {12,64}, {32,12}, {12,32}, {12,12}, {8,8}, {7,7}, {6,6}, {6,2}, {5,5}, {5,4}, {5,3},
		{5,2}, {4,4}, {2,4}, {4,5}, {4,2}, {4,3}, {4,2}, {3,3}, {3,4}, {1,32}, {32,1}, {30,1}, {20,1}, {1,20}, {7,1}, {6,1}, {5,1}, {4,1}, {2,3}, {1,4}, {3,1}, {1,3}, {2,2}, {2,1}, {1,2}, {1,1}
	}
	self.map.boxList = {}
	self.map.throwableBlockList = {}
	self.map.destroyableBlockList = {}
	self.projectiles = {}

	for currentTileType = 1, 99 do
		if currentTileType == 1 then
			for a = 1, #boxSizeList do
				local bw, bh = boxSizeList[a][1], boxSizeList[a][2]
				local count = bw * bh
				for x = 0, #tempColMap-bw+1 do
					for y = 0, #tempColMap[1]-bh+1 do
						local _count = 0
						for bx = x, x + (bw-1) do
							for by = y, y + (bh-1) do
								if not markedGrid[bx][by] then
									if tempColMap[bx][by] == currentTileType then
										_count = _count + 1
									end
								end
							end
						end
						if _count == count then
							local n = #self.map.boxList+1
							local wallType = currentTileType
							self.map.boxList[n] = { x = x * self.map.tileSize, y = y * self.map.tileSize, w = bw * self.map.tileSize, h = bh * self.map.tileSize, wallType = wallType, kind = "wall", id = n, solid = true }

							for bx = x, x + (bw-1) do
								for by = y, y + (bh-1) do
									markedGrid[bx][by] = true
								end
							end
						end
					end
				end
			end
		elseif currentTileType == 2 then
			local altBoxSizeList = {}
			for bsl = 1, #boxSizeList do
				if boxSizeList[bsl][2] == 1 then
					altBoxSizeList[#altBoxSizeList+1] = { boxSizeList[bsl][1], boxSizeList[bsl][2] }
				end
			end
			for a = 1, #altBoxSizeList do
				local bw, bh = altBoxSizeList[a][1], altBoxSizeList[a][2]
				local count = bw * bh
				for x = 0, #tempColMap-bw+1 do
					for y = 0, #tempColMap[1]-bh+1 do
						local _count = 0
						for bx = x, x + (bw-1) do
							for by = y, y + (bh-1) do
								if not markedGrid[bx][by] then
									if tempColMap[bx][by] == currentTileType then
										_count = _count + 1
									end
								end
							end
						end
						if _count == count then
							local n = #self.map.boxList+1
							local wallType = currentTileType
							self.map.boxList[n] = { x = x * self.map.tileSize, y = y * self.map.tileSize, w = bw * self.map.tileSize, h = bh * self.map.tileSize, wallType = wallType, kind = "wall", solidTop = true, id = n }

							for bx = x, x + (bw-1) do
								for by = y, y + (bh-1) do
									markedGrid[bx][by] = true
								end
							end
						end
					end
				end
			end
		elseif currentTileType == 3 then
			for x = 0, #tempColMap-1 do
				for y = 0, #tempColMap[1]-1 do
					if self.map.grid[x+1][y+1] == 3 then
						local n = "block" .. math.random(0,1000000)
						self.map.throwableBlockList[n] = { x = x * self.map.tileSize, y = y * self.map.tileSize, w = 1 * self.map.tileSize, h = 1 * self.map.tileSize, wallType = 3, kind = "block", canPickUp = true, id = n, solid = true }
					end
				end
			end
		elseif currentTileType == 4 then
			for x = 0, #tempColMap-1 do
				for y = 0, #tempColMap[1] do
					if self.map.grid[x+1][y+1] == 4 then
						local n = "dWall_" .. x+1 .. "_" .. y+1
						print(n)
						self.map.destroyableBlockList[n] = { x = x * self.map.tileSize, y = y * self.map.tileSize, w = 1 * self.map.tileSize, h = 1 * self.map.tileSize, wallType = 4, kind = "destroyableWall", id = n, solid = true }
					end
				end
			end
		elseif currentTileType == 5 then
			for x = 0, #tempColMap-1 do
				for y = 0, #tempColMap[1]-1 do
					if self.map.grid[x+1][y+1] == 5 then
						local n = "sand_" .. x+1 .. "_" .. y+1
						self.map.boxList[n] = { x = x * self.map.tileSize, y = y * self.map.tileSize, w = 1 * self.map.tileSize, h = 1 * self.map.tileSize, wallType = 5, kind = "sand", id = n, solidTop = true }
					end
				end
			end
		elseif currentTileType == 6 then
			for a = 1, #boxSizeList do
				local bw, bh = boxSizeList[a][1], boxSizeList[a][2]
				local count = bw * bh
				for x = 0, #tempColMap-bw+1 do
					for y = 0, #tempColMap[1]-bh+1 do
						local _count = 0
						for bx = x, x + (bw-1) do
							for by = y, y + (bh-1) do
								if not markedGrid[bx][by] then
									if tempColMap[bx][by] == currentTileType then
										_count = _count + 1
									end
								end
							end
						end
						if _count == count then
							local n = #self.map.boxList+1
							local wallType = currentTileType
							self.map.boxList[n] = { x = x * self.map.tileSize, y = y * self.map.tileSize, w = bw * self.map.tileSize, h = bh * self.map.tileSize, wallType = 6, kind = "wall", solid = true, slippery = true, id = n }

							for bx = x, x + (bw-1) do
								for by = y, y + (bh-1) do
									markedGrid[bx][by] = true
								end
							end
						end
					end
				end
			end
		elseif currentTileType == 17 or currentTileType == 18 then
			local altBoxSizeList = {}
			for bsl = 1, #boxSizeList do
				if boxSizeList[bsl][2] == 1 then
					altBoxSizeList[#altBoxSizeList+1] = { boxSizeList[bsl][1], boxSizeList[bsl][2] }
				end
			end
			for a = 1, #altBoxSizeList do
				local bw, bh = altBoxSizeList[a][1], altBoxSizeList[a][2]
				local count = bw * bh
				for x = 0, #tempColMap-bw+1 do
					for y = 0, #tempColMap[1]-bh+1 do
						local _count = 0
						for bx = x, x + (bw-1) do
							for by = y, y + (bh-1) do
								if not markedGrid[bx][by] then
									if tempColMap[bx][by] == currentTileType then
										_count = _count + 1
									end
								end
							end
						end
						if _count == count then
							local dir = 0
							if currentTileType == 17 then
								dir = 4
							elseif currentTileType == 18 then
								dir = -4
							end
							local n = #self.map.boxList+1
							self.map.boxList[n] = { x = x * self.map.tileSize, y = y * self.map.tileSize, w = bw * self.map.tileSize, h = bh * self.map.tileSize, wallType = currentTileType, kind = "wall", solidTop = true, conveyorDirection = dir, id = n }

							for bx = x, x + (bw-1) do
								for by = y, y + (bh-1) do
									markedGrid[bx][by] = true
								end
							end
						end
					end
				end
			end
		elseif currentTileType == 9 then
			for x = 0, #tempColMap-1 do
				for y = 0, #tempColMap[1]-1 do
					if (self.map.grid[x+1][y+1] == 9 and self.map.grid[x+1][y] == 0) or (self.map.grid[x+1][y+1] == 9 and self.map.grid[x][y+1] ~= 0 and self.map.grid[x+2][y+1] ~= 0) then
						local n = #self.map.boxList+1
						self.map.boxList[n] = { x = x * self.map.tileSize, y = y * self.map.tileSize, w = 1 * self.map.tileSize, h = 1 * self.map.tileSize, ladderTop = true, wallType = 2, kind = "wall", id = n }
					end
				end
			end
		end
	end

	for i, b in pairs(self.map.boxList) do
		self.map.collisionMap:add(self.map.boxList[i], self.map.boxList[i].x, self.map.boxList[i].y, self.map.boxList[i].w, self.map.boxList[i].h)
	end

	for i, b in pairs(self.map.throwableBlockList) do
		self.map.collisionMap:add(b, b.x, b.y, b.w, b.h)
	end

	for i, b in pairs(self.map.destroyableBlockList) do
		self.map.collisionMap:add(self.map.destroyableBlockList[i], b.x, b.y, b.w, b.h)
	end

	self.entities = {}
	for i, p in pairs(self.map.platforms) do
		local x, y, w, h = p.x * 16, p.y * 16, p.w * 16, p.h * 16
		self.entities[i] = platform:create { x = x, y = y, w = w, h = h, xSpeed = p.xSpeed, ySpeed = p.ySpeed, solid = p.solid, solidTop = p.solidTop, slippery = p.slippery, nameID = i, path = p.path, active = p.active, activateOnTouch = p.activateOnTouch, sinkWhenHeavy = p.sinkWhenHeavy }
		self.map.collisionMap:add(self.entities[i], x, y, w, h)
	end
	self.map.platforms = nil

	self:createPlayer()
	self:createEnemy()

	love.resize()

	win.setTitle("A loving tribute to Super Mario Bros. 2")

	lgr.setBackgroundColor(10,30,60)
end

function game:update(dt)
	local projectilesToRemove = {}

	for i, p in pairs(self.entities) do
		p:update(dt)
	end

	-- for i, e in pairs(self.enemies) do
	-- 	e:update(dt)
	-- end

	self.player:update(dt)

	for i, p in pairs(self.projectiles) do
		p:update(dt)
		if p.isDead then
			projectilesToRemove[#projectilesToRemove+1] = i
		end
	end

	for i, p in pairs(projectilesToRemove) do
		-- self.map.collisionMap:remove(self.projectiles[p])
		print("Destroying:", p)
		if self.projectiles[p].destroy then
			self.projectiles[p]:destroy()
		end
		self.projectiles[p] = nil
	end
end

function game:draw()
	self.canvas:clear(0,0,0,0)
	lgr.setCanvas(self.canvas)

	lgr.push()
	lgr.translate(-16,-16)
	for x = 1, self.map.width do
		for y = 1, self.map.height do
			if self.map.plantGrid[x][y].item ~= "" then
				lgr.setColor(0,200,0)
				lgr.rectangle("fill", x * 16 + 4, y * 16 - 4, 8, 4)
				-- lgr.rectangle("fill", x * 16, y * 16, 16, 16)
			end

			if (self.map.grid[x][y] > 0 and self.map.grid[x][y] < 3) or self.map.grid[x][y] == 5 or self.map.grid[x][y] == 6 or self.map.grid[x][y] == 17 or self.map.grid[x][y] == 18 then
				lgr.push()
				lgr.translate(x * 16, y * 16)
				lgr.setColor(255,255,255)
				if self.map.grid[x][y] == 3 then
					lgr.setColor(100,200,255)
				end
				lgr.rectangle("fill", 0, 0, 16, 16)
				lgr.pop()
			end

			if self.map.grid[x][y] == 9 then
				lgr.push()
				lgr.translate(x * 16, y * 16)
				lgr.setColor(160,160,160)
				lgr.rectangle("fill", 1, 0, 2, 16)
				lgr.rectangle("fill", 2, 3, 12, 2)
				lgr.rectangle("fill", 2, 11, 12, 2)
				lgr.rectangle("fill", 13, 0, 2, 16)
				lgr.pop()
			end
		end
	end

	lgr.pop()

	for i, b in pairs(self.map.boxList) do
		-- local b = self.map.boxList[i]
		if b.wallType == 1 then
			lgr.setColor(255,0,0,100)
		elseif b.wallType == 2 then
			lgr.setColor(0,255,0,100)
		elseif b.wallType == 5 then
			lgr.setColor(255,255,0,200)
		elseif b.wallType == 17 or b.wallType == 18 then
			lgr.setColor(0,0,0,200)
			if b.wallType == 17 then
				lgr.printf(">", b.x, b.y + 5, b.w, "center")
			else
				lgr.printf("<", b.x, b.y + 5, b.w, "center")
			end
			lgr.setColor(255,100,0,100)
		elseif b.wallType == 6 then
			lgr.setColor(0,200,255,200)
		else
			lgr.setColor(0,255,255,100)
		end
		lgr.rectangle("fill", b.x + 2, b.y + 2, b.w - 4, b.h - 4)
	end

	for i, b in pairs(self.map.throwableBlockList) do
		lgr.push()
		lgr.translate(0, 0)
		lgr.setColor(0,200,200)
		lgr.rectangle("fill", b.x, b.y, b.w, b.h / 2)
		lgr.setColor(0,255,255)
		lgr.rectangle("fill", b.x + 2, b.y + b.h / 2, b.w - 4, b.h / 2)
		lgr.pop()
	end

	for i, b in pairs(self.map.destroyableBlockList) do
		lgr.push()
		lgr.translate(0, 0)
		lgr.setColor(155,100,10)
		lgr.rectangle("fill", b.x, b.y, b.w, b.h)
		lgr.setColor(255,255,255,100)
		lgr.rectangle("fill", b.x, b.y, b.w-2, b.h-2)
		lgr.pop()
	end

	for i, p in pairs(self.entities) do
		p:draw()
	end

	-- for i, e in pairs(self.enemies) do
	-- 	e:draw()
	-- end

	self.player:draw()

	for i, p in pairs(self.projectiles) do
		p:draw()
	end

	lgr.push()
	lgr.translate(-16,-16)
	for x = 1, self.map.width do
		for y = 1, self.map.height do
			if self.map.waterMap[x][y] > 0 then
				lgr.push()
				lgr.translate(x * 16, y * 16)
				lgr.setColor(140,180,255,150)
				if self.map.waterMap[x][y-1] == 0 then
					local h = math.cos((tmr.getTime() * 5) + x / 5) * 1
					lgr.rectangle("fill", 0, 3 + h, 16, 13 - h)
				else
					lgr.rectangle("fill", 0, 0, 16, 16)
				end
				lgr.pop()
			end
		end
	end
	lgr.pop()

	if SHOWNUMBERS then
		lgr.setFont(font.dialog)
		for x = 1, self.map.width do
			for y = 1, self.map.height do
				if self.map.grid[x][y] > 0 then
					lgr.push()
					lgr.translate((x-1) * 16, (y-1) * 16 + 4)
					lgr.setColor(0,0,0)
					lgr.printf(self.map.grid[x][y], 1, 1, 16, "center")
					lgr.setColor(255,255,255)
					lgr.printf(self.map.grid[x][y], 0, 0, 16, "center")
					lgr.pop()
				end
			end
		end
	end

	local _, _, _, n = self.map.collisionMap:check(self.player)
	local t = "Player:\n"
	t = t .. "Collisions: " .. n .. "\n"
	t = t .. "Health: " .. self.player.health .. "\n"
 	t = t .. "X: " .. math.round(self.player.x, 4) .. "\n"
	t = t .. "Y: " .. math.round(self.player.y, 4) .. "\n"
	t = t .. "VX: " .. math.round(self.player.vx, 2) .. "\n"
	t = t .. "VY: " .. math.round(self.player.vy, 2) .. "\n"
	t = t .. "Run Mult: " .. tostring(self.player.runMult) .. "\n"
	t = t .. "Feet:  " .. tostring(self.player.feetl) .. "<" .. tostring(self.player.feetc) .. ">" .. tostring(self.player.feetr) .. "\n"
	t = t .. "Below: " .. tostring(self.player.tlb) .. "<" .. tostring(self.player.tcb) .. ">" .. tostring(self.player.trb) .. "\n"
	t = t .. "Jumping: " .. tostring(self.player.isJumping) .. "\n"
	t = t .. "Jump Height: " .. math.round(self.player.jumpHeight, 2) .. "\n"
	t = t .. "Jump Max: " .. self.player.jumpHeightMax .. "\n"
	t = t .. "On Ground: " .. tostring(self.player.onGround) .. "\n"
	t = t .. "On Block: " .. tostring(self.player.standingOnBlock) .. "\n"
	t = t .. "On Ladder: " .. tostring(self.player.onLadder) .. "\n"
	t = t .. "On Ice: " .. tostring(self.player.onIce) .. "\n"
	t = t .. "In Water: " .. tostring(self.player.inWater) .. "\n"
	t = t .. "Left Water: " .. tostring(self.player.leftWater) .. "\n"
	t = t .. "X, Y: " .. tostring(self.player.pickBlockX) .. ", " .. tostring(self.player.pickBlockY) .. "\n"
	t = t .. "Carrying:\n     " .. tostring(self.player.carryingBlockType) .. "\n     " .. tostring(self.player.carryingBlockID) .. "\n"
	t = t .. "Frame: " .. self.player.frame .. "\n"
	t = t .. "Pause Time: " .. self.player.pauseTime .. "\n"
	t = t .. "Platform: " .. self.player.attachedToPlatform .. "\n"

	lgr.setFont(font.tiny)
	lgr.setColor(255,255,255)
	lgr.print(t, 570, 4)

	lgr.setCanvas()
end

function game:keypressed(k, r)
	self.player:keypressed(k, false)

	if k == "f2" then
		SHOWNUMBERS = not SHOWNUMBERS
	end

	if k == "p" then
		stateStack:replace(game, { saveFile = 2 })
	end
end

function game:mousepressed(x, y, button)

end

function game:unload()
	print("Unloading Game State")
	self.loaded = false
	collectgarbage()
end

function game:blur(arg)
	local arg = arg or {}
	self.focused = false
end

function game:focus(arg)
	local arg = arg or {}
	self.focused = true
end

function game:resize(w, h)

end

function game:createPlayer()
	if self.player then
		self.map.collisionMap:remove(self.player)
	end
	self.player = smbPlayer:create { pState = game, x = 2 * self.map.tileSize, y = 14 * self.map.tileSize, iox = 2, ioy = 12,
					w = math.floor(0.7 * self.map.tileSize), h = math.floor(1.3 * self.map.tileSize), facing = 1,
					maxFallSpeed = 10, jumping = false, jumpVelocity = 0, world = self.map.collisionMap, carryingBlockType = 0, carryingBlockID = 0, kind == "player", id = "player" }
	self.map.collisionMap:add(self.player, self.player.x, self.player.y, self.player.w, self.player.h)
end

function game:createEnemy()
	-- self.entities = {}
	for i = 1, 5 do
		local id = "shyguy_" .. math.random(1,9999)
		self.entities[id] = shyguy:create { pState = game, x = (10 + i * 2) * self.map.tileSize, y = 18 * self.map.tileSize, iox = 2, ioy = 12,
						w = 12, h = 12, facing = 1,
						maxFallSpeed = 10, jumping = false, jumpVelocity = 0, world = self.map.collisionMap, carryingBlockType = 0, carryingBlockID = 0, kind == "enemy", nameID = id }
		self.map.collisionMap:add(self.entities[id], self.entities[id].x, self.entities[id].y, self.entities[id].w, self.entities[id].h)
	end
end

function game:createExplosion(ix, iy)
	print("BOOM!", ix, iy)
	ix = (ix + 4) / self.map.tileSize
	iy = (iy + 4) / self.map.tileSize
	for x = ix - 1, ix + 1 do
		for y = iy - 1, iy + 1 do
			local nx, ny = math.floor(x + 0.5), math.floor(y + 0.5)
			if self.map.grid[nx] then
				if self.map.grid[nx][ny] then
					if self.map.grid[nx][ny] == 4 then
						self.map.collisionMap:remove(self.map.destroyableBlockList["dWall_"..nx.."_"..ny])
						self.map.destroyableBlockList["dWall_"..nx.."_"..ny] = nil
						self.map.grid[nx][ny] = 0

						for jx = 0, 15, 4 do
							for jy = 0, 15, 4 do
								local n = "particle_" .. math.random(1,10000)
								game.projectiles[n] = smbParticle:create { x = ((nx-1) * 16) + jx, y = ((ny-1) * 16) + jy, vx = (jx * 4) - (8 * 4), vy = math.random(-30,30) ,
																			 color = {255-math.random(0,10),100,0}, life = 1, gravity = (math.random() * 50) + 50 }
							end
						end

					end
				end
			end
		end
	end
end

function game:checkIfBlockNextToWater(xx, yy)
	local b = {{xx-1, yy}, {xx+1, yy}, {xx, yy-1}, {xx, yy+1}}
	for i = 1, #b do
		local x, y = b[i][1], b[i][2]
		if self.map.waterMap[x][y] > 0 then
			self:gridFill(self.map.waterMap, x, y, 0, 1, function(...)
				local v = {...}
				local x, y = v[1], v[2]
				for jx = 0, 15, 4 do
					for jy = 0, 15, 4 do
						local n = "particle_" .. math.random(1,10000)
						local w = 100 * math.random()
						game.projectiles[n] = smbParticle:create { x = ((x-1) * 16) + jx + 2, y = ((y-1) * 16) + jy, vx = math.random(-2,2), vy = math.random(-20,-10),
																	 color = {w,w,255,255-100*math.random()}, life = 1, gravity = (math.random() * 50) + 50, shape = "square" }
					end
				end
			end)
			if yy > y then
				for j = 1, 15 do
					-- cron.after(j * 0.1-0.09, function()
						for k = 0, 15, 4 do
							local n = "particle_" .. math.random(1,10000)
							local w = 100 * math.random()
							game.projectiles[n] = smbParticle:create { x = (xx-1) * 16 + k + 2, y = (yy-1) * 16, vx = math.random(-2,2), vy = math.random(-20,-10),
																		 color = {w,w,255,255-100*math.random()}, life = 2, gravity = (math.random() * 50) + 11.5 * j, shape = "square" }
						end
					-- end)
				end
				print("Below")
			end
			print("YES", x, y)
		end
	end
end

-- 888888 88 88     88          dP""b8 88""Yb 88 8888b.
-- 88__   88 88     88         dP   `" 88__dP 88  8I  Yb
-- 88""   88 88  .o 88  .o     Yb  "88 88"Yb  88  8I  dY
-- 88     88 88ood8 88ood8      YboodP 88  Yb 88 8888Y"
function game:gridFill(grid,x ,y ,v ,oldv, func)
	if x < 0 or y < 0 or x > #grid or y > #grid[1] then return end
    if oldv == nil then oldv = grid[x][y] end
    if grid[x][y] == v then return end
    if grid[x][y] ~= oldv then return end
    grid[x][y] = v
    if func then func(x, y) end
    self:gridFill(grid,x-1,y,v,oldv, func)
    self:gridFill(grid,x+1,y,v,oldv, func)
    self:gridFill(grid,x,y+1,v,oldv, func)
    self:gridFill(grid,x,y-1,v,oldv, func)
end
